# -*- coding: utf-8 -*-
"""
Created on Fri Sep 11 13:08:02 2020
Deprecated Module !

@author: Daniel03

"""
from tkinter import *


class ButtonPerso(Canvas):
    
    def __init__(self, boss=None, **kwargs):
        Canvas.__init__(self,**kwargs)
        self.boss=boss
        self.border=kwargs.pop("borderwidth",3)
        self.height=kwargs.pop("height",600)
        self.width=kwargs.pop("width",300)
        self.command=kwargs.pop("command",None)
        for key, value in kwargs.items():
            setattr(self, key, value)
        
class CanvasPerso(Canvas):
    
    def __init__(self, boss=None, **kwargs):
        Canvas.__init__(self,borderwidth=borderwith, 
                        height=height, width=width, **kwargs)
        self.boss=boss
        self.border=kwargs.pop("borderwidth",3)
        self.height=kwargs.pop("height",600)
        self.width=kwargs.pop("width",300)
        
        for key, value in kwargs.items():
            setattr(self, key, value)
        
class FramePerso(Frame):
    
    def __init__(self, boss=None, **kwargs):
        Frame.__init__(self, boss,borderwidth=borderwidth,
                       height=height, width=width, **kwargs)
        self.boss=boss
        self.border=kwargs.pop("borderwidth",2)
        self.height=kwargs.pop("height",300)
        self.width=kwargs.pop("width",150)
        
        for key, value in kwargs.items():
            setattr(self, key, value)
        
        
        
if __name__=="__main__":
    
    root=Tk()
    root.title('AVGAPP')
    root.geometry("550x350+01+01")
    root.minsize("550","350")
    # root.iconbitmap("ag2dg-xripg-001.ico")
    
    fs=CanvasPerso(root,borderwidth=3,height=200,
                   width=300,relief="sunken" )
    fs.pack(expand=YES)
        
    
        








#!/usr/bin/env python

import pycsamt
import os 

# Check for setuptools package:

try:
    from setuptools import setup
except ImportError:
    setuptools = False
    from distutils.core import setup
else:
    setuptools = True

# LONG_DESCRIPTION = """
# pyCSAMT is a far field basic  open source software of controlled source audio-frequency magnetotellurics 
# for standard data processing , modeling and geophysical interpretation  enhancement.  
# """
with open(os.path.join(os.path.abspath('.'), 
                       'project_description.md'), 'r') as fm:
    # LONG_DESCRIPTION ="""{}""".format(
    #     ' '.join([descp for descp in fm.readlines() ]))
    LONG_DESCRIPTION =fm.read()

# The advantage of setuptools is that EXE wrappers are created on Windows,
# which allows Tab-completion for the script names from the system Scripts
# folder.
# Add names of scripts here. You can also specify the function to call
# by adding :func_name after the module name, and the name of the script
# can be customized before the equals sign. Actually some scripts may not work 
# because scripts is under designing with Tkinter  graphical interface 
# scripts were pseudocreated for waiting their implementations 

setup_kwargs = {}
setup_kwargs['entry_points'] = {'console_scripts': 
                    ['occam2d_build_in = pycsamt.gui.oc2d_bdin:main',
                     'write_avg2edi= pycsamt.gui.wa2edi:main',
                     'write_avg2j= pycsamt.gui.wa2j:main',
                     'corrected_edi = pycsamt.gui.corrected_edi:main',
                     'plot_model_oc2d = pycsamt.gui.p_moc2d:main',
                     'plot_pseudolog = pycsamt.gui.p2log:main',
                     'write_occam2oasis= pycsamt.gui.oas2f:main',
                     'write_occam2golden= pycsamt.gui.gs2f:main'
                     'write_iter2dat = pycsamt.gui.wi2d:main',
                     'write_drillhole= pycsamt.gui.cmake_dh:main']}
                     
                        

# But many people will not have setuptools installed, so we need to handle
# the default Python installation, which only has Distutils:

if setuptools is False:
    # Different script specification style for ordinary Distutils:

    setup_kwargs['scripts'] = [
        s.split(' = ')[1].replace('.', '/').split(':')[0] + '.py' for s in 
        setup_kwargs['entry_points']['console_scripts']]
    del setup_kwargs['entry_points']

    # "You must explicitly list all packages in packages: the Distutils will not
    # recursively scan your source tree looking for any directory with an
    # __init__.py file"

setup_kwargs['packages'] = [ 
                            'pycsamt',
                            'pycsamt.etc',
                            'pycsamt.ff',
                            'pycsamt.ff.core',
                            'pycsamt.gui',
                            'pycsamt.ff.processing',
                            'pycsamt.geodrill',
                            'pycsamt.geodrill.geoCore',
                            'pycsamt.geodrill.geoDB',
                            'pycsamt.geodrill.geoDB.sql_utils',
                            'pycsamt.geodrill.geoDB.sql_utils.sql_DB',
                            'pycsamt.viewer',
                            'pycsamt.modeling',
                            'pycsamt.utils',
                            ]
# force install mtpy . Once mtpy is installed , pyyaml and pyproj 
# should already installed too. 
     
setup_kwargs['install_requires'] = ['numpy>=1.8.1',
                                     'scipy>=0.14.0',
                                     'matplotlib',
                                     'mtpy >=1.1.0',
                                     'pyyaml',
                                     'pyproj',
                                     'configparser']
                                     
setup_kwargs['python_requires'] ='>=3.7'

authors =["Kouadio K. Laurent, ", 'Rong Liu, ', 
          'Binbin Mi, ','Chum-ning Liu, ', 'Albert O. Malory.']
authors_emails =['etanoyau@gmail.com,', 'liurongkaoyang@126.com,',
                'mibinbin@zju.edu.cn,', 'lifuming001@163.com,','amalory@zju.edu.cn']
setup(
	name="pycsamt",
	version=pycsamt.__version__,
	author=' '.join([aa for aa in authors]),
    author_email='kkouao@zju.edu.cn',
    maintainer="Kouadio K. Laurent",
    maintainer_email='etanoyau@gmail.com',
	description="A Python open-source toolkit for Controlled Source Audio-frequency Magnetotelluric ",
	long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/WEgeophysics/pyCSAMT",
    project_urls={
        "API Documentation"  : "https://pycsamt.readthedocs.io/en/latest/",
        "Home page" : "https://github.com/WEgeophysics/pyCSAMT/wiki",
        "Bugs tracker": "https://github.com/WEgeophysics/pyCSAMT/issues",
        "Installation guide" : "https://github.com/WEgeophysics/pyCSAMT/wiki/pyCSAMT-installation-guide-for-Windows--and-Linux", 
        "User guide" : "https://github.com/WEgeophysics/pyCSAMT/blob/develop/docs/pyCSAMT%20User%20Guide.pdf",
        },
	#data_files=[('', ['pycsamt/utils/epsg.npy',]),], #this will install datafiles in wearied palce such as ~/.local/
	include_package_data=True,
	license="GNU LESSER GENERAL PUBLIC LICENSE v3",
	classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        ],
	**setup_kwargs)



# -*- coding: utf-8 -*-
"""
===============================================================================
    Copyright © 2021  Kouadio K.Laurent
    
    This file is part of pyCSAMT.
    
    pyCSAMT is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    pyCSAMT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with pyCSAMT.  If not, see <https://www.gnu.org/licenses/>.

===============================================================================

.. _module-j:: `pycsamt.ff.core.j

   :synopsis: Deal with J files.  The J class can read and write an A.G.Jones file
             file, the 'standard format' of magnetotellurics.  Each section
             of the .edi file is given its own class, so the elements of each
             section are attributes for easy access.
             ...

Created on Thu Dec  3 22:31:16 2020

@author: @Daniel03
"""
import os , re, warnings
import datetime, shutil
import webbrowser
import numpy as np 


import pycsamt.ff.core.cs  as cs_obj
from pycsamt.etc.infos import notion

from pycsamt.ff.core.edi import Edi, Hmeasurement, Emeasurement 
from pycsamt.etc.infos import _sensitive as SB 
import pycsamt.utils.func_utils as func
from pycsamt.utils import exceptions as CSex
from pycsamt.utils._csamtpylog import csamtpylog

_logger = csamtpylog.get_csamtpy_logger(__name__)

class J_collection : 
    """
    J collection Class . Collect jfiles
    
    Arguments 
    ---------
        **list_of_jfiles** : list 
                list of Jfiles collection 
        **survey_name** : str 
                path to A.G. Jones Jfiles 
                
    Holds the following information:
        
    ================  ==========  =============================================
    Attributes         Type        Explanation
    ================  ==========  =============================================
    azimuth           ndarray,1     array_of all stations azimuth (deg) 
    latitude          ndarray,1     stations latitudes in degre decimal 
    longitude         ndarray,1     stations longitudes in degree decimal 
    elevation         ndarray,1     stations elevations in meter.
    period            list          list of array of collection periods.
                                     collect  all periods  from all stations . 
    app_rho           list          list of arrays of apparent resistivities
                                    (ohm.m)
    phase             list          list of array of phases in degree from 
                                    allstations.
    rhomax            list          list of  (rho + 1 standard error) arrays
    rhomin            list          list of (rho - 1 standard error) arrays
    phamax            list          list of (pha + 1 standard error )arrays
    phamin            list          list of (pha - 1 standard error) arrays 
    wrho              list          list of (weight for apparent_resistivity) 
                                    arrays  
    wpha              list          list of (weight for phase ) arrays 
    real              list          list of real part of the transfer
                                    function arrays 
    imag              list          list of imaginary part of the transfer
                                    function
    error             list          list of standard error  from alla stations 
    weight            list          list of  weight (Note: some weights are 
                                    in error!)
    ================  ==========  =============================================
    
    ====================  =====================================================
    Method                          Description 
    ====================  =====================================================
    collect_jfiles          read collections jfiles and populates attributes 
    plot_Topo_Stn_Azim      method to plot Topograpy , station seperation and 
                            Azimuth . 
    rewrite_jfiles          rewrite jfiles. 
    ====================  =====================================================
    
    """
    
    def __init__(self, list_of_jfiles =None , survey_name=None,  **kwargs): 
        self._logging =csamtpylog.get_csamtpy_logger(self.__class__.__name__)
        
        self.jfiles_list =list_of_jfiles
        self.survey_name =survey_name
        
        self.J=J()
        self.Location =cs_obj.Location ()
        self.Site =cs_obj.Site()
        
        self._latitude =None 
        self._longitude =None
        self._azimuth =None 
        self._elevation =None 
        self._station_names = None 
        
        self.nperiods = None 
        self.period =None 
        self.app_rho =None
        self.phase =None 
        
        self.rhomax = None 
        self.rhomin =None 
        self.phamax =None 
        self.phamin = None 
        self.wrho =None 
        self.wpha =None 
        
        self.real =None 
        self.imag =None 
        self.error =None 
        self.weight =None
        self.loc =None 
        self.id =None 
        

        if self.jfiles_list is not None :
            self.collect_jfiles()
        
    @property 
    def latitude(self): 
        return self.Location.latitude
    @latitude.setter 
    def latitude(self, latitude): 
        self.Location.latitude = latitude
        
    @property 
    def longitude(self): 
        return self.Location.longitude
    @longitude.setter 
    def longitude(self, longitude): 
        self.Location.longitude= longitude
        
    @property 
    def azimuth(self): 
        return self._azimuth
    @azimuth.setter 
    def azimuth(self, azimuth): 
        self._azimuth =np.array([float(azz) for azz in azimuth])
        
  
    @property 
    def elevation(self): 
        return self._elevation
    @elevation.setter 
    def elevation(self, elevation): 
        self._elevation =np.array([float(azz) for azz in elevation])
        
    @property 
    def stnames(self):
        return self._station_names
    @stnames.setter 
    def stnames(self, jstnames):
        if isinstance(jstnames, np.ndarray): jstnames=jstnames.tolist()
        elif isinstance(jstnames,list) : 
            self._station_names = J.jname(number_of_sites=len(jstnames), 
                                        survey_name=self.survey_name)
        elif isinstance(jstnames, int) :
            self._station_names = J.jname(number_of_sites=jstnames, 
                                        survey_name=self.survey_name)
        else : 
            try : 
                self._station_names = J.jname(number_of_sites= int(jstnames), 
                                          survey_name=self.survey_name)
            except :
                raise CSex.pyCSAMTError_J(
                    'Stations names must be on list or the number'
                    ' of stations not <{0}>.'.format(type(jstnames)))
    

    def j2edi(self, jfn=None, savepath =None, **kwargs): 
        """
        Method to convert j-files to edi files. Method calls CSAMT class object
        and get from this class edi infos 
        
        :param jfn: collection of jfiles or path-like str  
        :type list_of_files: str  

        :rtype: str
        :returns: edifiles from Jobjects
        
        :Example:
            
            >>> from pycsamt.ff.core.j import J_collection as JObjs
            >>> path2j = 'data/j' 
            >>> jObjs= JObjs().j2edi(path2j)
            
    
        """
        prospect =kwargs.pop('contractor_name', None)
        #hardwareInfos = kwargs.pop('hardware_name', None)
        fileby =kwargs.pop('fileby', 'jediSoftware')
        acqby =kwargs.pop('acqby', 'jedi')
        county =kwargs.pop('county', None)
        project_name =kwargs.pop('project_name', None)
        dipole_length =kwargs.pop('dipole_length', 100.)
        rotation_angle =kwargs.pop ('rotation_angle', 0.)
       
        if jfn is not None:
            self.jfiles_list = jfn

        if self.jfiles_list is None : 
            raise CSex.pyCSAMTError_J('No files found !'
                                      ' Please provide A.G. J-files ')
        # export to savepath 
        if savepath is None : # create a folder in your current work directory
            try :
                savepath = os.path.join(os.getcwd(), '_outputJ2EDI_')
                if not os.path.isdir(savepath):
                    os.mkdir(savepath)#  mode =0o666)
            except : 
                warnings.warn("It seems the path already exists !")
        
        # call CSAMT obj 
        csamt_jobj= cs_obj.CSAMT(data_fn=jfn)
        
        # create edi-obj and set attributes 
        for ii, stn in enumerate (csamt_jobj.station):
            # create an edi_obj and fill attributes 
            edi_obj=Edi()
            
            # fill Head info 
            edi_obj.Head.dataid= stn # ii mean for all the list 
            edi_obj.Head.acqby = acqby 
       
            if csamt_jobj.jsites_infos is None  or\
                csamt_jobj.jsites_infos==[None]: 
                csamt_jobj.jsites_infos = datetime.datetime.fromtimestamp(
                    os.stat(jfn).st_ctime) # return the creation date of file 
                
            edi_obj.Head.acqdate = csamt_jobj.jsites_infos 
            edi_obj.Head.fileby = fileby 
            edi_obj.Head.filedate = datetime.datetime.now(
                ).strftime('%m-%d-%Y %H:%M:%S')
            
            # get the name of location
            edi_obj.Head.loc = os.path.basename(csamt_jobj._fn)[:-4]

            if prospect is None : prospect ='MTnet'
            setattr(edi_obj, 'Head.prospect', prospect)
    
            if county is not None : county ='MT'
            setattr(edi_obj,'Head.county', county)
            
    
            edi_obj.Head.lat = csamt_jobj.lat[ii]
            edi_obj.Head.long = csamt_jobj.lon[ii] 
            edi_obj.Head.elev = round(csamt_jobj.elev[ii], 2)
        
            edi_obj.Head.maxsect =1000
            
            if project_name is None :
                project_name= os.path.basename(csamt_jobj._fn) 

            #=====>  set EDI OBJ INFOS
            # edi_obj.Info.maxinfo = 999
  
            edi_obj.Info.Source.__setattr__('project', project_name)
            edi_obj.Info.Source.__setattr__('survey',  edi_obj.Head.dataid)
            edi_obj.Info.Source.__setattr__('sitename', edi_obj.Head.dataid)
            edi_obj.Info.Processing.__setattr__('processedby', 'pyCSAMT' )
            edi_obj.Info.Processing.ProcessingSoftware.__setattr__(
                'name', edi_obj.Head.fileby )
    
           
            #====> definemeas 
            edi_obj.DefineMeasurement.maxchan =4
            edi_obj.DefineMeasurement.maxrun = len(csamt_jobj.station)
            edi_obj.DefineMeasurement.__setattr__('reftype' ,'CARTesien')
            edi_obj.DefineMeasurement.__setattr__('reflat',edi_obj.Head.lat  ) 
            edi_obj.DefineMeasurement.__setattr__('reflong', edi_obj.Head.long) 
            edi_obj.DefineMeasurement.__setattr__('refelev',round(edi_obj.Head.elev,2))
            
            #creating xxmeas object 
            codeID_dec = '{0}'.format((ii+1)/edi_obj.Head.maxsect)
            # codeID=  '{0:04}{1}'.format(ii * 10 + 1 , codeID_dec[1:] )
            edi_obj.DefineMeasurement.__setattr__('meas_ex', 
                                                  Emeasurement(**{
                                                      'id':'{0:04}{1}'.format(ii * 10 + 1 , 
                                                       codeID_dec[1:] ), 
                                                        'chtype':'EX', 
                                                        'x': -(dipole_length/2), 
                                                        'y': 0.,
                                                        'x2':dipole_length/2 , 
                                                        'y2':0, 
                                                        }))

            edi_obj.DefineMeasurement.__setattr__('meas_ey',
                                                  Emeasurement(**{
                                                      'id':'{0:04}{1}'.format(ii * 10 + 2 , 
                                                         codeID_dec[1:]), 
                                                        'chtype':'EY', 
                                                        'x':0., 
                                                        'y': -(dipole_length/2),
                                                        'x2':0., 
                                                        'y2':dipole_length/2 , 
                                                            }))
                                                                          

            edi_obj.DefineMeasurement.__setattr__('meas_hx', 
                                                  Hmeasurement(**{
                                                      'id':'{0:04}{1}'.format(ii * 10 + 3 , 
                                                       codeID_dec[1:] ), 
                                                        'chtype':'HX', 
                                                        'x':0., 
                                                        'y': 0.,
                                                        'x2':0., 
                                                        'y2':0. , 
                                                        }))

            edi_obj.DefineMeasurement.__setattr__('meas_hy',
                                                  Hmeasurement(**{
                                                      'id':'{0:04}{1}'.format(ii * 10 + 4 ,
                                                       codeID_dec[1:]), 
                                                    'chtype':'HY', 
                                                    'x':0., 
                                                    'y': 0.,
                                                    'x2':0., 
                                                    'y2':0. , 
                                                          }))
     
                     
            #====> EMAPSECT
            edi_obj.MTEMAP.sectid = stn 
            edi_obj.MTEMAP.__setattr__('nfreq', len(csamt_jobj.freq))
            edi_obj.MTEMAP.__setattr__(
                'ex', '{0:04}{1}'.format(ii * 10 + 1 , codeID_dec[1:] ))
            edi_obj.MTEMAP.__setattr__(
                'ey', '{0:04}{1}'.format(ii * 10 + 2 , codeID_dec[1:] ))
            edi_obj.MTEMAP.__setattr__(
                'hx', '{0:04}{1}'.format(ii * 10 + 3 , codeID_dec[1:] ))
            edi_obj.MTEMAP.__setattr__(
                'hy', '{0:04}{1}'.format(ii * 10 + 4 , codeID_dec[1:] ))
            
            #Frequency blocks , impendance and resistivity blocs 
            edi_obj.Z.freq = csamt_jobj.freq 
            #add rotation angle 
            edi_obj.Z.rotation_angle = rotation_angle
    
            
            # set phase and resistitivity including error propagation 
            # compute error propagation  
            #-->  initialize ndarray(nfreq, 2, 2) 
            res_array = np.zeros((edi_obj.Z.freq.size, 2,2 ), dtype = np.float)
            res_array_err = np.zeros((edi_obj.Z.freq.size, 2,2 ), dtype = np.float)
            phs_array = np.zeros((edi_obj.Z.freq.size, 2,2 ), dtype = np.float)
            phs_array_err = np.zeros((edi_obj.Z.freq.size, 2,2 ), dtype = np.float)
            
            #dictionnary of components . we set only component into XY . 
            res_array [:, 0 , 1 ] = csamt_jobj.resistivity[stn]  
            res_array_err [:, 0 , 1] = csamt_jobj.resistivity[stn] 
            phs_array[: , 0, 1] = csamt_jobj.phase[stn] 
            phs_array_err  [:, 0, 1]  = csamt_jobj.phase_err[stn] 
        
            #---> Recomputing z with resistivities- phase by using propagrations errors 
            edi_obj.Z.set_res_phase(res_array = res_array, phase_array=phs_array, 
                                    freq=  edi_obj.Z.freq, 
                                    res_err_array=res_array_err,
                                    phase_err_array=phs_array_err)
            
            edi_obj.write_edifile(savepath = savepath)
                
        if len(csamt_jobj.station) > 1: 
            print('-'*77)    
            print('---> {0} wrote sucessfully from j-files.\n---> see path:<{1}> '.\
                  format(len(csamt_jobj.station), savepath))
            print('-'*77)
                    
    
        
    def collect_jfiles (self, list_of_jfiles=None, jpath = None):
        """
        Collect the Jfile from jlist from jpath.Read and populate attributes .
        
        :param list_of_jfiles: list of jfiles
        :type list_of_jfiles: list 
        """
        if list_of_jfiles is None and jpath is not None : 
            self.jfiles_list = [os.path.join(jpath, jfile) 
                                for jfile in os.listdir (jpath)
                                if os.path.isfile(jfile) ==True]
        
        if list_of_jfiles is not None : 
            if isinstance( list_of_jfiles, str) : #we assume to read only one  file 
                self.jfiles_list = [list_of_jfiles] 
            self.jfiles_list = sorted(list_of_jfiles)
            
        elif self.jfiles_list is None : 
            raise CSex.pyCSAMTError_J ('Can not find a list of j files.'
                                       ' Please check your path !')
            
        # we assume that only file is read than put on list  for looping.
        if isinstance(self.jfiles_list, str): 
            self.jfiles_list =[self.jfiles_list]
        
        self._logging.info(
            'Number of J_files collected is %s', len(self.jfiles_list))
        
        if self.jfiles_list is not None : 
            jstations , lon,lat, azim, app_rho, pha, rhomax, rhomin, elev, period, \
            phamax, phamin , wrho, wpha, real , imag , error, weight = \
                [[] for ii in range(18)]
            
            # _set info_dic 
            sinfo, exl, eyl, exaz , eyaz , hxaz ,\
                cohmin,cohmax, maxt, nty, weig =[[] for ii in range (11)] 
 
            for jfile in self.jfiles_list : 

                self.J.read_j(j_fn =jfile)
                #----> comments blocks 
                if self.J.read_commentblock is True : 
                    sinfo.append(self.J.jinfo.site_infos),
                    exl.append(self.J.jinfo.ex_length)
                    eyl.append(self.J.jinfo.ey_length)
                    exaz.append(self.J.jinfo.ex_azimuth)
                    eyaz.append(self.J.jinfo.ey_azimuth)
                    cohmin.append(self.J.jinfo.coherence_min)
                    cohmax.append(self.J.jinfo.coherence_max) 
                    maxt.append(self.J.jinfo.max_type), 
                    nty.append(self.J.jinfo.ntype)
                    weig.append(self.J.jinfo.weight)
                
                #--- > information blocks 
                jstations.append(self.J._jnames)
                elev.append(self.J.jelev),
                period.append(self.J.jperiod)
                lon.append(self.J.jlon) 
                lat.append(self.J.jlat)
                azim.append(self.J.jazim)
                app_rho.append(self.J.japp_rho)
                pha.append(self.J.jpha)
                rhomax.append(self.J.jrhomax)
                rhomin.append(self.J.jrhomin), 
                phamax.append(self.J.jphamax)
                phamin.append(self.J.jphamin)
                wrho.append(self.J.jwrho)
                wpha.append(self.J.jwpha)
                real.append(self.J.jreal)
                imag.append(self.J.jimag)
                error.append(self.J.jerror)
                weight.append(self.J.jweight)
        
        
        self.stnames , self.elevation = jstations , elev
        self.longitude , self.latitude , self.azimuth = lon ,  lat, azim  
        self.period = period
        self.app_rho, self.rhomax, self.rhomin =  app_rho ,rhomax, rhomin 
        self.phase , self.phamax , self.phamin = pha , phamax , phamin 
        self.wrho , self.wpha =wrho , wpha
        self.weight , self.real , self.imag , \
            self.error = weight , real, imag , error
        
        #-----> set the code name of the site by calling site 
        if self.stnames is not None : 
            self.Site.stn_name = len(self.stnames)
            self.id = self.Site.stn_name 
        
        # set_info
        if self.J.read_commentblock: 
            
            self.J.jinfo.site_infos, self.J.jinfo.ex_length ,\
                self.J.jinfo.ey_length = sinfo , exl, eyl
            self.J.jinfo.ex_azimuth, self.J.jinfo.ey_azimuth , \
                self.J.jinfo.hx_azimuth = exaz, eyaz, hxaz 
            self.J.jinfo.coherence_max, self.J.jinfo.coherence_min ,\
                self.J.jinfo.max_type = cohmax, cohmin, maxt
            self.J.jinfo.ntype, self.J.jinfo.weight = nty, weig
                
            self.J.jinfo.jconfig ={
                'Site info': self.J.jinfo.site_infos, 
                'Ex length': self.J.jinfo.ex_length, 
                'Ey length':self.J.jinfo.ey_length, 
                'Ex azimuth': self.J.jinfo.ex_azimuth, 
                'Ey azimuth': self.J.jinfo.ey_azimuth, 
                'Hx azimuth':self.J.jinfo.hx_azimuth, 
                            'Coherence minimum':self.J.jinfo.coherence_min, 
                            'Coherence maximum': self.J.jinfo.coherence_max, 
                            'Maxtype':self.J.jinfo.max_type, 
                            'Ntype':self.J.jinfo.ntype , 
                            'Weight':self.J.jinfo.weight,
                            }
        
    def plot_Topo_Stn_Azim (self, list_of_jfiles =None, plot ='*',
                            show_stations =False , compute_azimuth =True , 
                savefig =None): 
        """
        plot Topography , Stn _Azim from Jfiles

        :param list_of_jfiles: list of jfiles 
        :type list_of_jfiles: list 
        
        :param plot: type of plot '*' means all the 03 plots or 
                      use `topo`, `stn`, `azim` to plot 
                     individually. *default* is '*'
        :type plot: str or int 
        
        :param show_stations: see the stations names as xlabel .
        :type show_stations: bool  
        
        :param compute_azimuth: if add azimuth, set azimuth to False 
                                *Default* is True 
        :type  compute_azimuth: bool  
        
        :param savefig: PathLike - path to save your figure
        :type savefig: str 
        
        .. note:: 
            Work but not stable ...
        """
        
        if list_of_jfiles is not None : self.jfiles_list= list_of_jfiles 
        if self.jfiles_list is None : raise CSex.pyCSAMTError_J(
                'Can not compute NoneType.Check your path')
        elif self.jfiles_list is not None :
            self.collect_jfiles(list_of_jfiles =self.jfiles_list )
            
        from pycsamt.viewer.plot import Plot1d 
        #--> create Plot1d Obj 
        plot_obj= Plot1d()
        #--> get easting northing arrays after converting from lat, lon
        jeasting , jnorthing = self.Location.get_eastnorth_array_from_latlon(
            arr_lat=self.latitude,  arr_lon = self.longitude)
                                                                           

        # ----> compute station separation using Profile obj 
        jstn_separation = cs_obj.Profile().stn_separation(
            easting =jeasting , northing =jnorthing)[0]
       
        #---- > interpolate Jstn_separation  so to get the same size as stations.
        from   scipy.interpolate import interp1d 
        xx_old =np.arange(jstn_separation.size)
        ff =interp1d(x=xx_old, y=jstn_separation, fill_value ='extrapolate')
        xx_new=np.arange(jstn_separation.size+1) # --> add one to get the same size 

        jstn_separation_extrap = ff(xx_new)

        
        if compute_azimuth is True :

            self.azimuth =func.compute_azimuth(
                easting =jeasting, northing=jnorthing , extrapolate =True)
                                                
            # self.Location.azimuth = np.concatenate((jeasting, jnorthing), axis=1)
            # self.azimuth = self.Location.azimuth 

        #---> copmute dipole length if not given 
        dipole_length =jstn_separation.mean()

        station_pk = np.apply_along_axis(
            lambda xx: xx * dipole_length, 0, np.arange(len(self.stnames)))

        #---> call plot func from plot1D
        plot_obj.plot_topo_sep_azim(plot=plot,
                                    elevation = self.elevation,
                                    station_names= self.stnames ,
                                    station_pk = station_pk , 
                                    azimuth=self.azimuth, 
                                    stn_separation = jstn_separation_extrap, 
                                    set_station_names=show_stations , 
                                    savepath =savefig )
                   

    def rewrite_jfiles (self, list_of_jfiles=None , savepath =None, 
                       survey_name =None ,  j_extension='.dat' ):
        """
        Method to rewrite A.G. johnson file (J-Format file).
        
        .. seealso::  http://mtnet.info/docs/jformat.txt
        
        :param list_of_files:  collection of Jfiles 
        :type list_of_files: list 
        
        :param survey_name:  name of exploration area 
        :type survey_name:   str  
        
        :param j_extension:  output format,
                            *Default* is '.dat'
        :type j_extension: str  
        
        :param savepath: output directory .If None ,
                        file will be store in current work directory
        :type savepath: str   
        """
        
        if list_of_jfiles is not None : self.jfiles_list =list_of_jfiles 
        if survey_name is not None : self.survey_name =survey_name 
        if self.jfiles_list is  None : 
            raise CSex.pyCSAMTError_J(
                'No files found to read . '
                 'Please check your path <%s>'% os.getcwd())  
        elif self.jfiles_list is not None : 
            self.collect_jfiles(list_of_jfiles =self.jfiles_list )
         
            
        #--- > start writing 
        write_jlines =[]
        if self.survey_name is None :
            code_name = 'ks{0:02}-{1}'
        elif self.survey_name is not None :
            code_name = self.survey_name[:-3]+'{0:02}-{1}'
        codespace= 5*' '
        
        for ii in range (len(self.id)) : 
            #-- > write Head j 
            write_jlines.append(''.join(
                [self.J._comment_mark,'{0:} :'.format(self.J.jinfo.progvers) , 
                code_name.format(ii, self.id[ii]), codespace, 
                datetime.datetime.now().strftime(self.J.jfmtime), 
                codespace, 'RECS'])+'\n')
            
            #---> write comments line if original file provided it.
            try : 
                if self.J.jinfo.site_infos is not None : write_jlines.append(
                        ''.join([self.J_comment_mark,'{0}:'.format('Site info'),
                                 self.J.jinfo.site_infos,'\n'] ))                                                                  
                if self.J.jinfo.ex_length is not None :
                    write_jlines.append(''.join([self.J_comment_mark,
                                                 '{0}:'.format('Ex length'),
                                                 self.J.jinfo.ex_length,'\n'] )) 
                if self.J.jinfo.ey_length is not None : 
                    write_jlines.append(''.join([self.J_comment_mark,
                                                 '{0}:'.format('Ey length'), 
                                                 self.J.jinfo.ey_length,'\n'] ))
                if self.J.jinfo.ex_azimuth is not None : 
                    write_jlines.append(''.join([self.J_comment_mark,
                                                 '{0}:'.format('Ex azimuth'), 
                                                 self.J.jinfo.ex_azimuth,'\n'] ))                                                                                                                                                   
                if self.J.jinfo.ey_azimuth is not None :
                    write_jlines.append(''.join([self.J_comment_mark,
                                                 '{0}:'.format('Ey azimuth'),
                                                 self.J.jinfo.ey_azimuth,'\n'] ) )                                                                    
                if self.J.jinfo.hx_azimuth is not None : 
                    write_jlines.append(''.join([self.J_comment_mark,
                                                 '{0}:'.format('Hx azimuth'),
                                                 self.J.jinfo.hx_azimuth,'\n'] ))                                                                                                                                   
                if self.J.jinfo.coherence_min is not None :
                    write_jlines.append(''.join([self.J_comment_mark,
                                                 '{0}:'.format('Coherence minimum'),
                                                 self.J.jinfo.coherence_min,'\n']))                                                                                                                                
                if self.J.jinfo.coherence_max is not None :
                    write_jlines.append(''.join([self.J_comment_mark,
                                                 '{0}:'.format('Coherence maximum'),
                                                 self.J.jinfo.coherence_max,'\n'] ))                                                                                                                                  
                if self.J.jinfo.max_type is not None : 
                    write_jlines.append(''.join([self.J_comment_mark,
                                                 '{0}:'.format('Maxtype'), 
                                                 self.J.jinfo.max_type,'\n'] ) )                                                            
                if self.J.jinfo.ntpype is not None :
                    write_jlines.append(''.join([self.J_comment_mark,
                                                 '{0}:'.format('Ntype'), 
                                                 self.J.jinfo.ntpype,'\n'] ) )                                                             
                if self.J.jinfo.weight is not None :
                    write_jlines.append(''.join([self.J_comment_mark,
                                                 '{0}:'.format('Ntype'),
                                                 self.J.jinfo.weight,'\n'] ) )  
            except : pass 
            
            #--- > write jlabels 
            write_jlines.append(''.join(['{0:<13}'.format(
                self.J._jlabels[0]),'=', "{0:>13.1f}".format(
                    self.azimuth[ii]), '\n'] ))
            write_jlines.append(''.join(['{0:<13}'.format(
                self.J._jlabels[1]),'=', "{0:>13.5f}".format(
                    self.latitude[ii]), '\n'] ))
            write_jlines.append(''.join(['{0:<13}'.format(
                self.J._jlabels[2]),'=', "{0:>13.5f}".format(
                    self.longitude[ii]), '\n'] ))
            write_jlines.append(''.join(['{0:<13}'.format(
                self.J._jlabels[3]),'=', "{0:>13.1f}".format(
                    self.elevation[ii]), '\n'] ))
            #---> write jnames , components  and nperiods 
            write_jlines.append(''.join(['{0}'.format(
                self.stnames[ii]), '{0:>7.1f}'.format(
                    self.azimuth[ii]),'\n']))
            write_jlines.append('{0}'.format(
                self.J.jmode)+'\n')
            write_jlines.append(' {0}'.format(
                self.J.jperiod.size)+'\n')
            # now  write value :

            for jj in range(self.J.jperiod.size): 
                if re.match(r'^Z+', self.J.jmode) is None or \
                    re.match(r'^T+', self.J.jmode) is None:
                    write_jlines.append (''.join( [
                        '{0:<12.3e}'.format(self.period[ii][jj]), 
                        '{0:^12.3e}'.format(self.app_rho [ii][jj]), 
                        '{0:^12.2e}'.format(self.phase [ii][jj]), 
                        '{:^12.3e}'.format(self.rhomax[ii][jj]), 
                        '{:^12.3e}'.format(self.rhomin [ii][jj]), 
                        '{:^12.2e}'.format(self.phamax[ii][jj]), 
                        '{:^12.2e}'.format(self.phamin[ii][jj]),
                        '{:^7.2f}'.format(self.wrho[ii][jj]), 
                        '{:^7.2f}'.format(self.wpha[ii][jj]),
                                                  ]))
                elif  re.match(r'^Z+', self.J.jmode) is not None or\
                    re.match(r'^T+', self.J.jmode) is not None:
                
                    write_jlines.append (''.join( [
                        '{0:<12.4e}'.format(self.period[ii][jj]), 
                        '{0:^12.4e}'.format(self.real [ii][jj]), 
                        '{0:^12.4e}'.format(self.imag [ii][jj]), 
                        '{:^12.3e}'.format(self.error[ii][jj]), 
                        '{:^12.2f}'.format(self.weight [ii][jj]), 
                                          ]))
                
                write_jlines.append('\n')
                
            with open('{0}{1}'.format(self.id[ii],j_extension), 'w', 
                      encoding='utf8') as fj:
                fj.writelines(write_jlines)
            
                
            if savepath is not None : 
                shutil.move ('{0}{1}'.format(
                    self.id[ii],j_extension), savepath)
                                                           
            write_jlines=[]  
            
        print('-'*77) 
        if savepath is None :savepath =os.getcwd()
        print('---> {0} J-files have been rewritten to <{1}>'
              ' <----'.format(len(self.id), savepath))
        print('-'*77)
                                                            

class J:
    """
    Class deal with A.G. Jones j-file. see : http://mtnet.info/docs/jformat.txt
    
    Arguments 
    ----------
            **j_fn** : str 
                    path to A.G. Jones Jfile . 
            
    Holds the following information:
        
    ==============  ==============  ===========================================
    Attributes         Type         Explanation
    ==============  ==============  ===========================================
    jazim           float           station azimuth in degree . 
    jlat            float           station latitude in degre decimal 
    jlon            float           station longitude in degree decimal 
    jelev           float           station elevation in meter.
    jmode           str             polarization mode.Polarization mode will 
                                    be use to compute jproperties.
    jnperiod        int             number of period . User can provide.If none 
                                    it will be computed automatically                   
    jperiod         (ndarray,1)     array of periods (s-1) or(1/second)1
    japp_rho        (ndarray,1)     array of apparent resistivities (ohm.m)
    jphase          (ndarray,1)     array of phase in degree 
    jrhomax         (ndarray,1)     rho + 1 standard error
    jrhomin         (ndarray,1)     rho - 1 standard error
    jphamax         (ndarray,1)     pha + 1 standard error 
    jphamin         (ndarray,1)     pha - 1 standard error
    jwrho           (ndarray,1)     weight for apparent_resistivity 
    jwpha           (ndarray,1)     weight for phase 
    jreal           (ndarray,1)     real part of the transfer function
    jimag           (ndarray,1)     imaginary part of the transfer function
    jerror          (ndarray ,1)    standard error 
    jweight         (ndarray,1)     weight (Note: some weights are in error!)
    ==============  ==============  ===========================================
    
    ====================  =====================================================
    Method                          Description 
    ====================  =====================================================
    jname                   staticmethod to generate A.G.Jones station codename
    jMode                   method for comformited H-Polarization and E-Polari-
                            zation mode . if mode is provided , il will check 
                            conformities, if not provied, Default is 'RXY'.
    read_j                  read the jfiles. 
    ====================  =====================================================
    
    .. note ::`-999` means missing data marker
            
    More attributes can be added by inputing a key word dictionary
    
    :Example:
        
        >>> from pycsamt.ff.core import J 
        >>> j=J()
        >>> jmode = j.jMode(polarization_type='RXY')
        >>>  print(jmode)
    """

    jdalpha , jdbeta ={'R':'apparent resistivities and phases', 
                          'S':'as "R" but upward-biased estimates', 
                          'Z':'impedances',
                          'Q':'as "Z" but upward biased estimates', 
                          'C':'"Z" expressed as Schmucker C function', 
                          'T':'GDS transfer functions'}, {'XX': 'xx MT impedance element', 
                                                            'XY':'xy MT impedance element', 
                                                            'YX':'yx MT impedance element',
                                                            'YY':' yy MT impedance element',
                                                            'TE':'TE MT impedance element', 
                                                            'TM':'TM MT impedance element', 
                                                            'AV':'Berdichevsky average', 
                                                            'DE':'determinant average', 
                                                            'ZX':'Tzx transfer function', 
                                                            'ZY':'Tzy transfer function'}

    def __init__(self, j_fn =None , **kws):
        
        self._logging=csamtpylog.get_csamtpy_logger(self.__class__.__name__)

        self.jfn =j_fn 
        
        self._jnames= None 
        self.jlon=None 
        self.jlat=None 
        self.jazim=None 
        self.jelev=None 
        self._jmode =None
        self.jnperiod=None
        self.jinfo =J_infos()

        
        self._jperiod=None
        self._japp_rho =None 
        self._jpha =None 
        self._jrhomax=None 
        self._jrhomin=None 
        self._jphamax =None 
        self._jphamin=None 
        self._jwrho =None  
        self._jwpha =None 
        
        
        self._jreal =None
        self._jimag =None 
        self._jerror =None 
        self._jweight =None 
        self._jhead =None 

        self._jnan = -999

        self._comment_mark='#'
        self.jfmtime ='%m-%d-%Y %H:%M:%S'
        self._jlabels =['>AZIMUTH','>LATITUDE', 
                        '>LONGITUDE', '>ELEVATION' ]
        
        self.read_commentblock =False  
        
        for jkey in list (kws.keys()): 
            setattr(self, jkey, kws[jkey])
        
    #----set jproperties and jsetter-functions -----
        
    @property 
    def jperiod (self): 
        return self._jperiod 
    @jperiod.setter 
    def jperiod(self, jperds): 
        for ii, item in enumerate(jperds): 
            if item in ['','*'] or item == str(self._jnan): 
                jperds[ii]=np.nan
        try :
            self._jperiod=np.array([float(ii) for ii in jperds])
        except : 
            raise CSex.pyCSAMTError_J(
                'Can not convert "str"jperiod value to float.')
    
    @property 
    def japp_rho (self): 
        return self._app_rho 
    @japp_rho.setter 
    def japp_rho (self, japp_rho): 
        for ii, item in enumerate(japp_rho): 
            if item in ['','*'] or item == str(self._jnan): 
                japp_rho[ii]=np.nan
        try : 
            self._app_rho=np.array([float(ii) for ii in japp_rho])
        except :
            raise CSex.pyCSAMTError_J(
                'Can not convert "str" apparent resistivity value to float.')
    
    @property 
    def jpha(self): 
        return self._jpha 
    @jpha.setter 
    def jpha (self, jphase): 
        for ii, item in enumerate(jphase): 
            if item in ['','*'] or item == str(self._jnan):
                jphase[ii]=np.nan
        try : 
            self._jpha=np.array([float(ii) for ii in jphase])
        except : 
            raise CSex.pyCSAMTError_J(
                'Could not convert "str" phase value to float.')
        
    @property 
    def jrhomax(self): 
        return self._jrhomax 
    @jrhomax.setter 
    def jrhomax (self, jrhomax): 
        for ii, item in enumerate(jrhomax): 
            if item in ['','*']or item == str(self._jnan): 
                jrhomax[ii]=np.nan
        try : 
            self._jrhomax=np.array([float(ii) for ii in jrhomax])
        except : 
            raise CSex.pyCSAMTError_J(
                'Could not convert "str"jrhomax value to float.')
        
    @property 
    def jphamax(self): 
        return self._jphamax 
    @jphamax.setter 
    def jphamax (self, jphamax): 
        for ii, item in enumerate(jphamax): 
            if item in  ['','*'] or item == str(self._jnan): 
                jphamax[ii]=np.nan
        try : 
            self._jphamax=np.array([float(ii) for ii in jphamax])
        except :
            raise CSex.pyCSAMTError_J(
                'Could not convert "str"jphamax value to float.')
        
    @property 
    def jrhomin(self): 
        return self._jrhomin 
    @jrhomin.setter 
    def jrhomin (self, jrhomin): 
        for ii, item in enumerate(jrhomin): 
            if item in ['','*']or item == str(self._jnan):
                jrhomin[ii]=np.nan
        try : 
            self._jrhomin=np.array([float(ii) for ii in jrhomin])
        except : 
            raise CSex.pyCSAMTError_J(
                'Could not convert "str"jrhomin value to float.')
        
    @property 
    def jphamin(self): 
        return self._jphamin 
    @jphamin.setter 
    def jphamin (self, jphamin): 
        for ii, item in enumerate(jphamin): 
            if item in ['','*'] or item == str(self._jnan):
                jphamin[ii]=np.nan
        try :
            self._jphamin=np.array([float(ii) for ii in jphamin])
        except : 
            raise CSex.pyCSAMTError_J(
                'Could not convert "str" jphamin value to float.')
        
    @property 
    def jwrho(self): 
        return self._jwrho
    @jwrho.setter 
    def jwrho (self, jwrho): 
        for ii, item in enumerate(jwrho): 
            if item in  ['','*'] or item == str(self._jnan): 
                jwrho[ii]=np.nan
        try : 
            self._jwrho=np.array(
                [float(ii) for ii in jwrho])
        except : 
            raise CSex.pyCSAMTError_J(
                'Could not convert "str" jwrho value to float.')
    
    @property 
    def jwpha(self): 
        return self._jwpha
    @jwpha.setter 
    def jwpha (self, jwpha): 
        for ii, item in enumerate(jwpha): 
            if item in ['','*'] or item == str(self._jnan): 
                jwpha[ii]=np.nan
        try : 
            self._jwpha=np.array([float(ii) for ii in jwpha])
        except : 
            raise CSex.pyCSAMTError_J(
                'Could not convert "str" jwphase value to float.')
    
    @property 
    def jreal(self): 
        return self._jreal
    @jreal.setter 
    def jreal (self, jreal): 
        for ii, item in enumerate(jreal): 
            if item in ['','*'] or item == str(self._jnan): 
                jreal[ii]=np.nan
        try : 
            self._jreal=np.array([float(ii) for ii in jreal])
        except : 
            raise CSex.pyCSAMTError_J(
                'Could not convert "str" jreal value to float.')   


    @property 
    def jimag(self): 
        return self._jimag
    @jimag.setter 
    def jimag (self, jimag): 
        for ii, item in enumerate(jimag): 
            if item in ['','*'] or item == str(self._jnan):
                jimag[ii]=np.nan
        try : 
            self._jimag=np.array([float(ii) for ii in jimag])
        except :
            raise CSex.pyCSAMTError_J(
                'Could not convert "str" jimag value to float.')   

    @property 
    def jerror(self): 
        return self._jerror
    @jerror.setter 
    def jerror (self, jerror): 
        for ii, item in enumerate(jerror): 
            if item in ['','*'] or item == str(self._jnan):
                jerror[ii]=np.nan
        try : 
            self._jerror=np.array([float(ii) for ii in jerror])
        except :
            raise CSex.pyCSAMTError_J(
                'Could not convert "str" jerror value to float.')  

    @property 
    def jweight(self): 
        return self._jweight
    @jweight.setter 
    def jweight (self, jweight):
        if jweight is None : self._jweight = np.full((self.japp_rho.size,),1.)
        for ii, item in enumerate(jweight): 
            if item in ['','*'] or item == str(self._jnan): 
                jweight[ii]=np.nan
        try : 
            self._jweight=np.array([float(ii) for ii in jweight])
        except : 
            raise CSex.pyCSAMTError_J(
                'Could not convert "str" jweight value to float.')  


    @property 
    def jmode (self):
        return self._jmode 
    
    @jmode.setter 
    def jmode(self, jpolar):
        if not isinstance(jpolar, str):
            raise CSex.pyCSAMTError_J(
                'jMode polarization must be on str not <{}>'.format(type(jpolar)))
        self._jmode =self.jMode(polarization_type=jpolar)
        

    @staticmethod
    def jname( number_of_sites, survey_name= None,):
        """
        Staticmethod for generate alphanumeric `station name` (case sensitive 
        when reading A.G Jones files) survey XXX, station 001. 
        
        :param  number_of_sites: number of stations . 
        :type number_of_sites: int 
        
        :param survey_name: place location name . 
        :type survey_name: str 
        
        :returns: list of alphanumeric station names 
        :rtype: list
              
        """
        alpha, cunm ='abcdefghijklmnopqrstuvwxyz', []
        if not isinstance(number_of_sites, int): 
            try :number_of_sites = int(number_of_sites)
            except : 
                raise CSex.pyCSAMTError_J(
                    'Number of sites must be int not <{0}>.'.
                     format(type(number_of_sites)))
        for ss in range(number_of_sites):
            if survey_name is not None :
                symb=survey_name[:2].lower() 
            else:  symb='cs'
            if ss == 0 :
                cunm.append('{0}{1}000'.format(symb, alpha[0])) 
            elif (ss % 10  == 0): 
                cunm.append('{0}{1}{2}'.format(symb, alpha[ss//10], "000"))
            else :
                cunm.append('{0}{1}{2:03}'.format(symb, alpha[ss//10], ss%10))
        
        return cunm
    
   
    def jMode(self, polarization_type='RXY'):
        """
        Jmode is conformited a different set of H-Polarization ans E-Polarization. 
        for more detail :
            
        .. seealso:: http://mtnet.info/docs/jformat.txt
        
        :Convention:The convention used is for RXY to represent
                the E-polarization (TE) mode, and for RYX
                to represent the B-polarization mode.
        """
        comb =[ ('{0}{1}'.format(jakeys, jbkeys), '{0}/{1}'.format(
            javalues, jbvalues)) for jakeys, javalues  in self.jdalpha.items() 
               for jbkeys, jbvalues in self.jdbeta.items()]
            
        jmodeDICT = {jkeym:jvaluem for jkeym, jvaluem in zip \
                     ([pol for pol, info in comb],[info for pol, info in comb] )}

        if polarization_type not in list(jmodeDICT.keys()) :
            warnings.warn(
                'Polarization Type  provided <{0}> for JMode is unacceptable.'
                'Please consults the Dictionnary of jPolarisation mode above.')
            
            [print('{0:<5}:{1:<55}'.format(keys,values)) for keys,
             values in zip(list(jmodeDICT.keys()), list(jmodeDICT.values()))]
                         
            print(notion.j)
            
            raise CSex.pyCSAMTError_J(
                'Value provided is not in polarization mode '
                 '.Please consult the dict above.') 
        
        return polarization_type
    
    def read_j (self, j_fn=None  ): 
        """
        Method to read Jformat file.
        
        :param j_fn: path to jfile 
        :type j_fn: str
        """ 

        jdata=[]
        if j_fn is not None : self.jfn =j_fn 
        if self.jfn is None : 
            raise CSex.pyCSAMTError_J(
                'Error file. Please Provide the right path!')
  
        self._logging.info('Reading A.G.Jones J-format "%s"'%os.path.basename(
            self.jfn))
        
        if self.jfn is not None : 

            if SB.which_file(filename=self.jfn) =='j': 
                with open(self.jfn , 'r', encoding='utf8') as fj: 
                    j_data_lines =fj.readlines()
            else :
                self._logging.warn("File <%s>is not J-format file." %self.jfn)
                warnings.warn('File <%s> is not J-Format File.'% self.jfn)
                raise CSex.pyCSAMTError_J(
                    'File provided doesn no match the J-format.'
                    ' Please consult :"http://mtnet.info/docs/jformat.txt" {0}'.\
                     format(webbrowser.open('http://mtnet.info/docs/jformat.txt')))
            
            #set attribute when loop in the file . 
            for ss , jitems in enumerate(j_data_lines):
                # set comments infos from jfile :
                if re.match(r'^#', jitems) is not None : 

                    
                    j_comminfo, jcommval,*ignore =jitems.strip().split(':')
                    
                    if 'Site info'.lower() in j_comminfo.lower():
                        self.J.jinfo.site_infos = jcommval
                    elif 'Ex length'.lower() in j_comminfo.lower():
                        self.J.jinfo.ex_length = jcommval 
                    elif 'Ey length'.lower() in j_comminfo.lower():
                        self.J.jinfo.ey_length = jcommval
                    elif 'Ex azimuth'.lower() in j_comminfo.lower():
                        self.J.jinfo.ex_azimuth = jcommval 
                    elif 'Ey azimuth'.lower() in j_comminfo.lower():
                        self.J.jinfo.ey_azimuth = jcommval
                    
                    elif 'Hx azimuth'.lower() in j_comminfo.lower():
                        self.J.jinfo.hx_azimuth = jcommval
                    elif 'Coherence minimum'.lower() in j_comminfo.lower():
                        self.J.jinfo.coherence_min = jcommval
                    elif 'Coherence maximum'.lower() in j_comminfo.lower():
                        self.J.jinfo.coherence_max = jcommval
                    
                    elif 'Maxtype'.lower() in j_comminfo.lower():
                        self.J.jinfo.max_type = jcommval
                    elif 'Ntype'.lower() in j_comminfo.lower():
                        self.J.jinfo.ntype = jcommval
                    elif 'Weight'.lower() in j_comminfo.lower():
                        self.J.jinfo.weight = jcommval
                    
                    self.read_commentblock = True 
                #set longitude value 
                elif  re.match(r'^>', jitems) is not None :
                    jparam, jparamvalue =jitems.strip().split('=')
                    if 'azimuth'  in jparam.lower() :
                        self.jazim =jparamvalue 
                    elif 'longitude' in jparam.lower():
                        self.jlon=jparamvalue
                    elif 'latitude' in jparam.lower():
                        self.jlat =jparamvalue
                    elif 'elevation' in jparam.lower():
                        self.jelev =jparamvalue
                
                else :
                    #---> Get the station name , jmode and number of periods 
                    jmain = jitems.strip().split()
                    if jmain == None or jmain ==[]: pass 
                    elif len(jmain) < 5 :
                        try : 
                            self.jnperiods= float(jmain[0])
                        except :
                            if jmain[0] in SB._j :self.jmode=jmain[0]
                            else : self._jnames = jmain[0]
                            pass 
                        else :self.jnperiod = jmain[0] 
                    elif len(jmain) >= 5 and jmain !=['']:

                            jdata.append(np.array(jmain))

                
            JDAT= func.concat_array_from_list(list_of_array= jdata, concat_axis=0)

            if (re.match(r'^Z', self.jmode) is None) or \
                (re.match(r'^T', self.jmode)  is None) :
                if JDAT.shape[1] == 9: 
                    jDATA = [np.reshape(jitem,( jitem.shape[0],))\
                             for jitem in np.hsplit(JDAT,9) ]

                    self.jperiod, self.japp_rho, self.jpha, self.jrhomax , self.jrhomin , \
                        self.jphamax, self.jphamin , self.jwrho , self.jwpha = jDATA
                        
                else :
                    warnings.warn (
                        'J-FORMAT expects to get "9" records'
                         ' values like <{0}>'.format(*list(notion.j_recordR.keys())))
                    raise CSex.pyCSAMTError_J(
                        'For data type=R?? Only  9 Range values are not '
                        'acceptable. You provided <{0}>'.format(JDAT.shape[1]))
            elif (re.match(r'^Z', self.jmode) is True) or \
                (re.match(r'^T', self.jmode)  is True):
                if JDAT.shape[1] == 5 : 
                    jDATA = [np.reshape(ii,( ii.shape[0],)) \
                             for ii in np.hsplit(JDAT,5) ]
                    self.jperiod , self.jreal , self.jimag ,\
                        self.jerror, self.jweight = jDATA
                else :
                    warnings.warn('JFORMAT for GSC responses expects'
                                  ' to get "5" records values like<{0}>'.\
                                      format(*list(notion.j_recordZ.keys())))
                    raise CSex.pyCSAMTError_J(
                        'For data type=R?? Only  9 Range values'
                        ' are not acceptable. You provided <{0}>'.format(JDAT.shape[1]))
        
     
class J_infos (object):
    """
     J_infos class - set the information of J_file
    
    """

    def __init__(self, **kwargs): 
        
        self.progvers = 'WRITTEN BY pyCSAMT'
        self.site_infos = None 
        self.ex_length = None 
        self.ey_length =None 
        self.ex_azimuth =None 
        self.ey_azimuth =None 
        self.hx_azimuth =None
        self.coherence_min = None
        self.coherence_max =None 
        self.max_type =None 
        self.ntype =1
        self.weight ='Y'
        #---> use j config to quick write file .
        self.jconfig ={'Site info': self.site_infos, 
                       'Ex length': self.ex_length, 
                       'Ey length':self.ey_length, 
                       'Ex azimuth': self.ex_azimuth, 
                       'Ey azimuth': self.ey_azimuth, 
                       'Hx azimuth':self.hx_azimuth, 
                       'Coherence minimum':self.coherence_min, 
                       'Coherence maximum': self.coherence_max, 
                       'Maxtype':self.max_type, 
                       'Ntype':self.ntype , 
                       'Weight':self.weight,
                       }
        
        for keys in list(kwargs.keys()): 
            self.__setattr__(keys, kwargs[keys])

 
 
if __name__=='__main__': 
    jdata =r'./data/j/S00.dat'
    J_collection().j2edi(jfn=jdata, 
                         #rotation_angle=np.array([0.])
                         )
    
    
    
    
    
    
    
    
    
    
    
    
    



#!/usr/bin/env python

"""
===============================================================================
    Copyright © 2021  Kouadio K. Laurent
    
    This file is part of pyCSAMT.
    
    pyCSAMT is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    pyCSAMT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with pyCSAMT.  If not, see <https://www.gnu.org/licenses/>.

===============================================================================    

.. module:: Z
   :synopsis: Deal with MT responses Z and Tipper
        Module belongs to  MTpy .see  https://mtpy2.readthedocs.io/en/develop/ 
        CSAMT and MT are  very close geophysical methods
        We slighty edited this module for CSAMT purpose. 
        However we keep the Exception of MTpy for authorships.
        If you use this module , please you may cite authors:
        Kirkby, A.L., Zhang, F., Peacock, J., Hassan, R., Duan, J., 2019. 
        The MTPy software package for magnetotelluric data analysis and visualisation.
        Journal of Open Source Software, 4(37), 1358. https://doi.org/10.21105/joss.01358
        ...

.. moduleauthor:: Jared Peacock <jpeacock@usgs.gov>
.. moduleauthor:: Lars Krieger

.. edited ::  @Daniel03
"""

import cmath
import copy
import math

# =================================================================
import numpy as np

import pycsamt.ff.processing.mtcalculator as MTcc
import pycsamt.utils.exceptions as MTex
from pycsamt.utils._csamtpylog import csamtpylog

_logger = csamtpylog.get_csamtpy_logger(__name__)


# ==============================================================================
# Resistivity and phase object
# ==============================================================================
class ResPhase(object):
    """
    resistivity and phase container
    .. module:: Z
    .. moduleauthor:: Jared Peacock <jpeacock@usgs.gov>
    .. moduleauthor:: Lars Krieger
    """

    def __init__(self, z_array=None, z_err_array=None, freq=None, **kwargs):
        self._logger = csamtpylog.get_csamtpy_logger(self.__class__.__name__)

        self._z = z_array
        self._z_err = z_err_array

        self._resistivity = None
        self._phase = None

        self._resistivity_err = None
        self._phase_err = None

        self.freq = freq

        for key in kwargs:
            setattr(self, key, kwargs[key])

    @property
    def resistivity(self):
        return self._resistivity

    @resistivity.setter
    def resistivity(self, res_array):
        self._resistivity = res_array

    @property
    def resistivity_err(self):
        return self._resistivity_err

    @resistivity_err.setter
    def resistivity_err(self, res_err_array):
        self._resistivity_err = res_err_array

    @property
    def phase(self):
        return self._phase

    @phase.setter
    def phase(self, phase_array):
        self._phase = phase_array

    @property
    def phase_err(self):
        return self._phase_err

    @phase_err.setter
    def phase_err(self, phase_err_array):
        self._phase_err = phase_err_array

    def compute_resistivity_phase(self, z_array=None, z_err_array=None,
                                  freq=None):
        """
        compute resistivity and phase from z and z_err
        """

        if z_array is not None:
            self._z = z_array
        if z_err_array is not None:
            self._z_err = z_err_array
        if freq is not None:
            self.freq = freq

        #if self._z is None or self._z_err is None or self.freq is None: #The _z_err can be None!!!
        if self._z is None or self.freq is None:
            raise MT_Z_Error('Values are None, check _z, _z_err, freq')

        self._resistivity = np.apply_along_axis(lambda x: np.abs(x) ** 2 / self.freq * 0.2,
                                                0, self._z)
        self._phase = np.rad2deg(np.angle(self._z))

        self._resistivity_err = np.zeros_like(self._resistivity, dtype=np.float)
        self._phase_err = np.zeros_like(self._phase, dtype=np.float)

        # calculate resistivity and phase
        if self._z_err is not None:
            for idx_f in range(self.freq.size):
                for ii in range(2):
                    for jj in range(2):

                        r_err, phi_err = MTcc.z_error2r_phi_error(
                                self._z[idx_f, ii, jj].real,
                                self._z[idx_f, ii, jj].imag,
                                self._z_err[idx_f, ii, jj])
                        self._resistivity_err[idx_f, ii, jj] = \
                            self._resistivity[idx_f, ii, jj] * r_err

                        self._phase_err[idx_f, ii, jj] = phi_err

    def set_res_phase(self, res_array, phase_array, freq, res_err_array=None,
                      phase_err_array=None):
        """
        Set values for resistivity (res - in Ohm m) and phase
        (phase - in degrees), including error propagation.

        :param res_array: resistivity array in Ohm-m
        :type res_array: np.ndarray(num_freq, 2, 2)

        :param phase_array: phase array in degrees
        :type phase_array: np.ndarray(num_freq, 2, 2)

        :param freq: frequency array in Hz
        :type freq: np.ndarray(num_freq)

        :param res_err_array: resistivity error array in Ohm-m
        :type res_err_array: np.ndarray(num_freq, 2, 2)

        :param phase_err_array: phase error array in degrees
        :type phase_err_array: np.ndarray(num_freq, 2, 2)
        """

        self._logger.info ('Resetting z and z_err')

        self._resistivity = res_array
        self._phase = phase_array
        self.freq = freq
        self._resistivity_err = res_err_array
        self._phase_err = phase_err_array

        # assert real array:
        if np.linalg.norm(np.imag(res_array)) != 0:
            raise MTex.MTpyError_inputarguments('Error - array "res" is not' + \
                                                'real valued !')

        if np.linalg.norm(np.imag(phase_array)) != 0:
            raise MTex.MTpyError_inputarguments('Error - array "phase" is' + \
                                                'not real valued !')

        abs_z = np.sqrt(5.0 * self.freq * (self.resistivity.T)).T
        self._z = abs_z * np.exp(1j * np.radians(self.phase))

        self._z_err = np.zeros_like(self._z, dtype=np.float)
        # ---------------------------
        # error propagation:
        if self._resistivity_err is None or self._phase_err is None:
            return

        for idx_f in range(self.freq.shape[0]):
            for ii in range(2):
                for jj in range(2):
                    abs_z = np.sqrt(5 * self.freq[idx_f] * \
                                    self.resistivity[idx_f, ii, jj])
                    rel_error_res = self.resistivity_err[idx_f, ii, jj] / \
                                    self.resistivity[idx_f, ii, jj]
                    # relative error varies by a factor of 0.5, which is the
                    # exponent in the relation between them:
                    abs_z_error = 0.5 * abs_z * rel_error_res

                    self._z_err[idx_f, ii, jj] = max(MTcc.propagate_error_polar2rect(
                        abs_z,
                        abs_z_error,
                        self.phase[idx_f, ii, jj],
                        self.phase_err[idx_f, ii, jj]))

    @property
    def res_xx(self):
        return self._resistivity[:, 0, 0]

    @property
    def res_xy(self):
        return self._resistivity[:, 0, 1]

    @property
    def res_yx(self):
        return self._resistivity[:, 1, 0]

    @property
    def res_yy(self):
        return self._resistivity[:, 1, 1]

    @property
    def phase_xx(self):
        return self._phase[:, 0, 0]

    @property
    def phase_xy(self):
        return self._phase[:, 0, 1]

    @property
    def phase_yx(self):
        return self._phase[:, 1, 0]

    @property
    def phase_yy(self):
        return self._phase[:, 1, 1]

    @property
    def res_err_xx(self):
        return self._resistivity_err[:, 0, 0]

    @property
    def res_err_xy(self):
        return self._resistivity_err[:, 0, 1]

    @property
    def res_err_yx(self):
        return self._resistivity_err[:, 1, 0]

    @property
    def res_err_yy(self):
        return self._resistivity_err[:, 1, 1]

    @property
    def phase_err_xx(self):
        return self._phase_err[:, 0, 0]

    @property
    def phase_err_xy(self):
        return self._phase_err[:, 0, 1]

    @property
    def phase_err_yx(self):
        return self._phase_err[:, 1, 0]

    @property
    def phase_err_yy(self):
        return self._phase_err[:, 1, 1]

    # calculate determinant values
    @property
    def _zdet(self):
        return np.array([np.linalg.det(zz) ** .5 for zz in self._z])

    @property
    def _zdet_var(self):
        if self._z_err is not None:
            return np.array([abs(np.linalg.det(zzv)) ** .5 for zzv in self._z_err])
        else:
            return np.ones_like(self._zdet, dtype=np.float)

    @property
    def phase_det(self):
        return np.arctan2(self._zdet.imag, self._zdet.real) * (180 / np.pi)

    @property
    def phase_det_err(self):
        return np.arcsin(self._zdet_var / abs(self._zdet)) * (180 / np.pi)

    @property
    def res_det(self):
        return 0.2 * (1. / self.freq) * abs(self._zdet) ** 2

    @property
    def res_det_err(self):
        return 0.2 * (1. / self.freq) * np.abs(self._zdet + self._zdet_var) ** 2 - self.res_det


# ==============================================================================
# Impedance Tensor Class
# ==============================================================================
class Z(ResPhase):
    """
    Z class - generates an impedance tensor (Z) object.

    Z is a complex array of the form (n_freq, 2, 2),
    with indices in the following order:

        - Zxx: (0,0)
        - Zxy: (0,1)
        - Zyx: (1,0)
        - Zyy: (1,1)

    All errors are given as standard deviations (sqrt(VAR))

    :param z_array: array containing complex impedance values
    :type z_array: numpy.ndarray(n_freq, 2, 2)


    :param z_err_array: array containing error values (standard deviation)
                        of impedance tensor elements
    :type z_err_array: numpy.ndarray(n_freq, 2, 2)

    :param freq: array of frequency values corresponding to impedance tensor
                 elements.
    :type freq: np.ndarray(n_freq)


    =============== ===========================================================
    Attributes      Description
    =============== ===========================================================
    freq             array of frequencies corresponding to elements of z
    rotation_angle   angle of which data is rotated by
    z                impedance tensor
    z_err             estimated errors of impedance tensor
    resistivity      apparent resisitivity estimated from z in Ohm-m
    resistivity_err  apparent resisitivity error
    phase            impedance phase (deg)
    phase_err        error in impedance phase
    =============== ===========================================================

    =================== =======================================================
    Methods             Description
    =================== =======================================================
    det                  calculates determinant of z with errors
    invariants           calculates the invariants of z
    inverse              calculates the inverse of z
    remove_distortion    removes distortion given a distortion matrix
    remove_ss            removes static shift by assumin Z = S * Z_0
    norm                 calculates the norm of Z
    only1d               zeros diagonal components and computes
                         the absolute valued mean of the off-diagonal
                         components.
    only2d               zeros diagonal components
    res_phase            computes resistivity and phase
    rotate               rotates z positive clockwise, angle assumes
                         North is 0.
    set_res_phase        recalculates z and z_err, needs attribute freq
    skew                 calculates the invariant skew (off diagonal trace)
    trace                calculates the trace of z
    =================== =======================================================

    :Example: ::

        >>> import mtpy.core.z as mtz
        >>> import numpy as np
        >>> z_test = np.array([[0+0j, 1+1j], [-1-1j, 0+0j]])
        >>> z_object = mtz.Z(z_array=z_test, freq=[1])
        >>> z_object.rotate(45)
        >>> z_object.resistivity
    """

    def __init__(self, z_array=None, z_err_array=None, freq=None):
        """
        Initialise an instance of the Z class.

        :param z_array: array containing complex impedance values
        :type z_array: numpy.ndarray(n_freq, 2, 2)

        :param z_err_array: array containing error values (standard deviation)
                            of impedance tensor elements
        :type z_err_array: numpy.ndarray(n_freq, 2, 2)

        :param freq: array of frequency values corresponding to impedance
                     tensor elements.
        :type freq: np.ndarray(n_freq)

        Initialises the attributes with None
        """

        ResPhase.__init__(self)

        self._z = z_array
        self._z_err = z_err_array
        self._freq = freq

        if z_array is not None:
            if len(z_array.shape) == 2 and z_array.shape == (2, 2):
                if z_array.dtype in ['complex', 'float', 'int']:
                    self._z = np.zeros((1, 2, 2), 'complex')
                    self._z[0] = z_array

        if z_err_array is not None:
            if len(z_err_array.shape) == 2 and z_err_array.shape == (2, 2):
                if z_err_array.dtype in ['complex', 'float', 'int']:
                    self._z_err = np.zeros((1, 2, 2), 'complex')
                    self._z_err[0] = z_err_array

        self.rotation_angle = 0.
        if self._z is not None:
            self.rotation_angle = np.zeros((len(self._z)))

        if self._z is not None:
            self.compute_resistivity_phase()

    # ---frequency-------------------------------------------------------------
    @property
    def freq(self):
        """
        Frequencies for each impedance tensor element

        Units are Hz.
        """
        return self._freq

    @freq.setter
    def freq(self, freq_arr):
        """
        Set the array of freq.

        :param freq_arr: array of frequnecies (Hz)
        :type freq_arr: np.ndarray
        """

        if freq_arr is not None:
            self._freq = np.array(freq_arr)
        else:
            return None

        if self.z is not None:
            if len(self.z.shape) == 3:
                if self._freq.size != len(self.z):
                    self._logger.warn('length of freq list/array not correct'
                                      '({0} instead of {1})'.format(self._freq.size,
                                                                    len(self.z)))
                    return
                else:
                    try:
                        self.compute_resistivity_phase()
                    except IndexError:
                        print('Need to input frequency array')

    # ----impedance tensor -----------------------------------------------------
    @property
    def z(self):
        """
        Impedance tensor

        np.ndarray(nfreq, 2, 2)
        """
        return self._z

    @z.setter
    def z(self, z_array):
        """
        Set the attribute 'z'.


        :param z_array: complex impedance tensor array
        :type z_array: np.ndarray(nfreq, 2, 2)

        Test for shape, but no test for consistency!

        Nulling the rotation_angle
        """

        try:
            if len(z_array.shape) == 3 and z_array.shape[1:3] == (2, 2):
                if z_array.dtype in ['complex', 'float', 'int']:
                    self._z = z_array
        except IndexError:
            try:
                if len(z_array.shape) == 2 and z_array.shape == (2, 2):
                    if z_array.dtype in ['complex', 'float', 'int']:
                        self._z = np.zeros((1, 2, 2), 'complex')
                        self._z[0] = z_array
            except IndexError:
                self._logger.error('provided Z array does not have correct dimensions- Z unchanged')

        if isinstance(self.rotation_angle, float):
            self.rotation_angle = np.repeat(self.rotation_angle,
                                            len(self._z))

        # for consistency recalculate resistivity and phase
        if self._z is not None and self._z_err is not None:
            try:
                self.compute_resistivity_phase()
            except IndexError:
                self._logger.error('Need to input frequency array')

    # ----impedance error-----------------------------------------------------
    @property
    def z_err(self):
        return self._z_err

    @z_err.setter
    def z_err(self, z_err_array):
        """
        Set the attribute z_err

        :param z_err_array: error of impedance tensor array as standard
                            deviation
        :type z_err_array: np.ndarray(nfreq, 2, 2)
        """
        if z_err_array.shape != self.z.shape:
            self._logger.warn('z_err_array shape {0} is not same shape as z {1}'.format(
                z_err_array.shape, self.z.shape))
        self._z_err = z_err_array

        # for consistency recalculate resistivity and phase
        if self._z_err is not None and self._z is not None:
            try:
                self.compute_resistivity_phase()
            except IndexError:
                self._logger.error('Need to input frequency array')

    @property
    def inverse(self):
        """
            Return the inverse of Z.

            (no error propagation included yet)

        """

        if self.z is None:
            self._logger.warn('z array is "None" - I cannot invert that')
            return

        inverse = copy.copy(self.z)
        for idx_f in range(len(inverse)):
            try:
                inverse[idx_f, :, :] = np.array((np.matrix(self.z[idx_f, :, :])).I)
            except:
                raise MTex.MTpyError_Z('The {0}ith impedance'.format(idx_f + 1) + \
                                       'tensor cannot be inverted')

        return inverse

    def rotate(self, alpha):
        """
        Rotate Z array by angle alpha.

        Rotation angle must be given in degrees. All angles are referenced
        to geographic North, positive in clockwise direction.
        (Mathematically negative!)

        In non-rotated state, X refs to North and Y to East direction.

        Updates the attributes
            - *z*
            - *z_err*
            - *zrot*
            - *resistivity*
            - *phase*
            - *resistivity_err*
            - *phase_err*

        """

        if self.z is None:
            self._logger.warn('Z array is "None" - I cannot rotate that')
            return

        # check for iterable list/set of angles - if so, it must have length
        # 1 or same as len(tipper):
        if np.iterable(alpha) == 0:
            try:
                degreeangle = float(alpha % 360)
            except ValueError:
                self._logger.error('"Angle" must be a valid number (in degrees)')
                return

            # make an n long list of identical angles
            lo_angles = [degreeangle for ii in self.z]
        else:
            if len(alpha) == 1:
                try:
                    degreeangle = float(alpha % 360)
                except ValueError:
                    self._logger.error('"Angle" must be a valid number (in degrees)')
                    return
                # make an n long list of identical angles
                lo_angles = [degreeangle for ii in self.z]
            else:
                try:
                    lo_angles = [float(ii % 360) for ii in alpha]
                except ValueError:
                    self._logger.error('"Angles" must be valid numbers (in degrees)')
                    return

        self.rotation_angle = np.array([(oldangle + lo_angles[ii]) % 360
                                        for ii, oldangle in enumerate(self.rotation_angle)])

        if len(lo_angles) != len(self.z):
            self._logger.warn('Wrong number of "angles" - I need {0}'.format(len(self.z)))
            # self.rotation_angle = 0.
            return

        z_rot = copy.copy(self.z)
        z_err_rot = copy.copy(self.z_err)

        for idx_freq in range(len(self.z)):

            angle = lo_angles[idx_freq]
            if np.isnan(angle):
                angle = 0.

            if self.z_err is not None:
                z_rot[idx_freq], z_err_rot[idx_freq] = \
                    MTcc.rotatematrix_incl_errors(self.z[idx_freq, :, :],
                                                  angle,
                                                  self.z_err[idx_freq, :, :])
            else:
                z_rot[idx_freq], z_err_rot = \
                    MTcc.rotatematrix_incl_errors(self.z[idx_freq, :, :],
                                                  angle)

        self.z = z_rot
        if self.z_err is not None:
            self.z_err = z_err_rot

        # for consistency recalculate resistivity and phase
        self.compute_resistivity_phase()

    def remove_ss(self, reduce_res_factor_x=1., reduce_res_factor_y=1.):
        """
        Remove the static shift by providing the respective correction factors
        for the resistivity in the x and y components.
        (Factors can be determined by using the "Analysis" module for the
        impedance tensor)

        Assume the original observed tensor Z is built by a static shift S
        and an unperturbated "correct" Z0 :

             * Z = S * Z0

        therefore the correct Z will be :
            * Z0 = S^(-1) * Z


        :param reduce_res_factor_x: static shift factor to be applied to x
                                    components (ie z[:, 0, :]).  This is
                                    assumed to be in resistivity scale
        :type reduce_res_factor_x: float or iterable list or array

        :param reduce_res_factor_y: static shift factor to be applied to y
                                    components (ie z[:, 1, :]).  This is
                                    assumed to be in resistivity scale
        :type reduce_res_factor_y: float or iterable list or array

        :returns: static shift matrix,
        :rtype: np.ndarray ((2, 2))

        :returns: corrected Z
        :rtype: mtpy.core.z.Z

        .. note:: The factors are in resistivity scale, so the
                  entries of  the matrix "S" need to be given by their
                  square-roots!

        """

        # check for iterable list/set of reduce_res_factor_x - if so, it must
        # have length 1 or same as len(z):
        if np.iterable(reduce_res_factor_x) == 0:
            try:
                x_factor = float(reduce_res_factor_x)
            except ValueError:
                self._logger.error('reduce_res_factor_x must be a valid numbers')
                return

            lo_x_factors = np.repeat(x_factor, len(self.z))
        elif len(reduce_res_factor_x) == 1:
            try:
                x_factor = float(reduce_res_factor_x)
            except ValueError:
                self._logger.error('reduce_res_factor_x must be a valid numbers')
                return
            lo_x_factors = np.repeat(x_factor, len(self.z))
        else:
            try:
                lo_x_factors = np.repeat(x_factor,
                                         len(reduce_res_factor_x))
            except ValueError:
                self._logger.error('"reduce_res_factor_x" must be valid numbers')
                return

        if len(lo_x_factors) != len(self.z):
            self._logger.error('Wrong number Number of reduce_res_factor_x - need {0}'.format(len(self.z)))
            return

        # check for iterable list/set of reduce_res_factor_y - if so,
        # it must have length 1 or same as len(z):
        if np.iterable(reduce_res_factor_y) == 0:
            try:
                y_factor = float(reduce_res_factor_y)
            except ValueError:
                self._logger.error('"reduce_res_factor_y" must be a valid numbers')
                return

            lo_y_factors = np.repeat(y_factor, len(self.z))
        elif len(reduce_res_factor_y) == 1:
            try:
                y_factor = float(reduce_res_factor_y)
            except ValueError:
                self._logger.error('"reduce_res_factor_y" must be a valid numbers')
                return
            lo_y_factors = np.repeat(y_factor, len(self.z))
        else:
            try:
                lo_y_factors = np.repeat(y_factor,
                                         len(reduce_res_factor_y))
            except ValueError:
                self._logger.error('"reduce_res_factor_y" must be valid numbers')
                return

        if len(lo_y_factors) != len(self.z):
            self._logger.error('Wrong number Number of "reduce_res_factor_y"' + \
                               '- need {0} '.format(len(self.z)))
            return

        z_corrected = copy.copy(self.z)
        static_shift = np.zeros((len(self.z), 2, 2))

        for idx_f in range(len(self.z)):
            # correct for x-direction
            z_corrected[idx_f, 0, :] = self.z[idx_f, 0, :] / \
                                       np.sqrt(lo_x_factors[idx_f])
            # correct for y-direction
            z_corrected[idx_f, 1, :] = self.z[idx_f, 1, :] / \
                                       np.sqrt(lo_y_factors[idx_f])
            # make static shift array
            static_shift[idx_f, 0, 0] = np.sqrt(lo_x_factors[idx_f])
            static_shift[idx_f, 1, 1] = np.sqrt(lo_y_factors[idx_f])

        return static_shift, z_corrected

    def remove_distortion(self, distortion_tensor, distortion_err_tensor=None):
        """
        Remove distortion D form an observed impedance tensor Z to obtain
        the uperturbed "correct" Z0:
        Z = D * Z0

        Propagation of errors/uncertainties included


        :param distortion_tensor: real distortion tensor as a 2x2
        :type distortion_tensor: np.ndarray(2, 2, dtype=real)


        :param distortion_err_tensor: default is None
        :type distortion_err_tensor: np.ndarray(2, 2, dtype=real),

			:returns: input distortion tensor
        :rtype: np.ndarray(2, 2, dtype='real')

			:returns: impedance tensor with distorion removed
        :rtype: np.ndarray(num_freq, 2, 2, dtype='complex')


			:returns: impedance tensor error after distortion is removed
        :rtype: np.ndarray(num_freq, 2, 2, dtype='complex')


    		:Example: ::

    			>>> import mtpy.core.z as mtz
    			>>> distortion = np.array([[1.2, .5],[.35, 2.1]])
    			>>> d, new_z, new_z_err = z_obj.remove_distortion(distortion)
        """

        if distortion_err_tensor is None:
            distortion_err_tensor = np.zeros_like(distortion_tensor)
        # for all freq, calculate D.Inverse, then obtain Z0 = D.I * Z
        try:
            if not (len(distortion_tensor.shape) in [2, 3]) and \
                    (len(distortion_err_tensor.shape) in [2, 3]):
                raise ValueError('Shape not the same')
            if len(distortion_tensor.shape) == 3 or \
                            len(distortion_err_tensor.shape) == 3:
                self._logger.info('Distortion is not time-dependent - take only first' + \
                                  'of given distortion tensors')
                try:
                    distortion_tensor = distortion_tensor[0]
                    distortion_err_tensor = distortion_err_tensor[0]
                except IndexError:
                    raise ValueError('distortion tensor the wrong shape')

            if not (distortion_tensor.shape == (2, 2)) and \
                    (distortion_err_tensor.shape == (2, 2)):
                raise ValueError('Shape not the same')

            distortion_tensor = np.matrix(np.real(distortion_tensor))

        except ValueError:
            raise MTex.MTpyError_Z('The array provided is not a proper' + \
                                   'distortion tensor')

        try:
            DI = distortion_tensor.I
        except np.linalg.LinAlgError:
            raise MTex.MTpyError_Z('The provided distortion tensor is' + \
                                   'singular - I cannot invert that!')

        # propagation of errors (using 1-norm) - step 1 - inversion of D:
        DI_err = np.zeros_like(distortion_err_tensor)

        # todo :include error on  determinant!!
        # D_det = np.linalg.det(distortion_tensor)

        dummy, DI_err = MTcc.invertmatrix_incl_errors(distortion_tensor,
                                                      distortion_err_tensor)

        # propagation of errors - step 2 - product of D.inverse and Z;
        # D.I * Z, making it 4 summands for each component:
        z_corrected = np.zeros_like(self.z)
        z_corrected_err = np.zeros_like(self.z_err)

        for idx_f in range(len(self.z)):
            z_corrected[idx_f] = np.array(np.dot(DI, np.matrix(self.z[idx_f])))
            for ii in range(2):
                for jj in range(2):
                    z_corrected_err[idx_f, ii, jj] = np.sum(np.abs(
                        np.array([DI_err[ii, 0] * \
                                  self.z[idx_f, 0, jj], \
                                  DI[ii, 0] * \
                                  self.z_err[idx_f, 0, jj], \
                                  DI_err[ii, 1] * \
                                  self.z[idx_f, 1, jj], \
                                  DI[ii, 1] * \
                                  self.z_err[idx_f, 1, jj]])))

        return distortion_tensor, z_corrected, z_corrected_err

    def _compute_det_variance(self):
        """
        compute the variance of the determinant of Z, 
        """

    @property
    def only_1d(self):
        """
        Return Z in 1D form.

        If Z is not 1D per se, the diagonal elements are set to zero,
        the off-diagonal elements keep their signs, but their absolute
        is set to the mean of the original Z off-diagonal absolutes.
        """

        z1d = copy.copy(self.z)

        for ii in range(len(z1d)):
            z1d[ii, 0, 0] = 0
            z1d[ii, 1, 1] = 0
            sign01 = np.sign(z1d[ii, 0, 1])
            sign10 = np.sign(z1d[ii, 1, 0])
            mean1d = 0.5 * (z1d[ii, 1, 0] + z1d[ii, 0, 1])
            z1d[ii, 0, 1] = sign01 * mean1d
            z1d[ii, 1, 0] = sign10 * mean1d

        return z1d

    @property
    def only_2d(self):
        """
        Return Z in 2D form.

        If Z is not 2D per se, the diagonal elements are set to zero.
        """

        z2d = copy.copy(self.z)

        for ii in range(len(z2d)):
            z2d[ii, 0, 0] = 0
            z2d[ii, 1, 1] = 0

        return z2d

    @property
    def trace(self):
        """
        Return the trace of Z

        :returns: Trace(z)
        :rtype: np.ndarray(nfreq, 2, 2)

        """

        tr = np.array([np.trace(ii) for ii in self.z])

        return tr

    @property
    def trace_err(self):
        """
        Return the trace of Z

        :returns: Trace(z)
        :rtype: np.ndarray(nfreq, 2, 2)

        """

        tr_err = None
        if self.z_err is not None:
            tr_err = np.zeros_like(self.trace, dtype=np.float)
            tr_err[:] = self.z_err[:, 0, 0] + self.z_err[:, 1, 1]

        return tr_err

    @property
    def skew(self):
        """
        Returns the skew of Z as defined by Z[0, 1] + Z[1, 0]

        .. note:: This is not the MT skew, but simply the linear algebra skew


        :returns: skew
        :rtype: np.ndarray(nfreq, 2, 2)
        """

        skew = np.array([ii[0, 1] - ii[1, 0] for ii in self.z])

        return skew

    @property
    def skew_err(self):
        """
        Returns the skew error of Z as defined by Z_err[0, 1] + Z_err[1, 0]

        .. note:: This is not the MT skew, but simply the linear algebra skew

        :returns: skew_err
        :rtype: np.ndarray(nfreq, 2, 2)
        """

        skew_err = None
        if self.z_err is not None:
            skew_err = np.zeros_like(self.skew, dtype=np.float)
            skew_err[:] = self.z_err[:, 0, 1] + self.z_err[:, 1, 0]

        return skew_err

    @property
    def det(self):
        """
        Return the determinant of Z

        :returns: det_Z
        :rtype: np.ndarray(nfreq)
        """

        det_Z = np.array([np.linalg.det(ii) for ii in self.z])

        return det_Z

    @property
    def det_err(self):
        """
        Return the determinant of Z error

        :returns: det_Z_err
        :rtype: np.ndarray(nfreq)
        """
        det_Z_err = None
        if self.z_err is not None:
            det_Z_err = np.zeros_like(self.det, dtype=np.float)
            # components of the impedance tensor are not independent variables
            # so can't use standard error propagation
            # calculate manually:
            # difference of determinant of z + z_err and z - z_err then divide by 2
            det_Z_err[:] = np.abs(np.linalg.det(self.z + self.z_err) - \
                                  np.linalg.det(self.z - self.z_err)) / 2.
        # det_Z_err[:] = np.abs(self.z[:, 1, 1] * self.z_err[:, 0, 0]) +\
        #                           np.abs(self.z[:, 0, 0] * self.z_err[:, 1, 1]) +\
        #                           np.abs(self.z[:, 0, 1] * self.z_err[:, 1, 0]) +\
        #                           np.abs(self.z[:, 1, 0] * self.z_err[:, 0, 1])

        return det_Z_err

    @property
    def norm(self):
        """
        Return the 2-/Frobenius-norm of Z

        :returns: norm
        :rtype: np.ndarray(nfreq)
        """

        norm = np.array([np.linalg.norm(ii) for ii in self.z])

        return norm

    @property
    def norm_err(self):
        """
        Return the 2-/Frobenius-norm of Z  error

        :returns: norm_err
        :rtype: np.ndarray(nfreq)
        """
        norm_err = None

        if self.z_err is not None:
            norm_err = np.zeros_like(self.norm, dtype=np.float)
            for idx, z_tmp in enumerate(self.z):
                value = self.norm[idx]
                error_matrix = self.z_err[idx]
                radicand = 0.
                for ii in range(2):
                    for jj in range(2):
                        radicand += (error_matrix[ii, jj] * \
                                     np.real(z_tmp[ii, jj])) ** 2
                        radicand += (error_matrix[ii, jj] * \
                                     np.imag(z_tmp[ii, jj])) ** 2

                norm_err[idx] = 1. / value * np.sqrt(radicand)

        return norm_err

    @property
    def invariants(self):
        """
        Return a dictionary of Z-invariants.

        Contains
    		-----------
    			* z1
    			* det
    			* det_real
    			* det_imag
    			* trace
    			* skew
    			* norm
    			* lambda_plus/minus,
    			* sigma_plus/minus
        """

        invariants_dict = {}

        z1 = (self.z[:, 0, 1] - self.z[:, 1, 0]) / 2.
        invariants_dict['z1'] = z1

        invariants_dict['det'] = self.det[0]

        det_real = np.array([np.linalg.det(ii) for ii in np.real(self.z)])
        invariants_dict['det_real'] = det_real

        det_imag = np.array([np.linalg.det(ii) for ii in np.imag(self.z)])
        invariants_dict['det_imag'] = det_imag

        invariants_dict['trace'] = self.trace

        invariants_dict['skew'] = self.skew

        invariants_dict['norm'] = self.norm

        invariants_dict['lambda_plus'] = z1 + np.sqrt(z1 * z1 / self.det)

        invariants_dict['lambda_minus'] = z1 - np.sqrt(z1 * z1 / self.det)

        invariants_dict['sigma_plus'] = 0.5 * self.norm ** 2 + \
                                        np.sqrt(0.25 * self.norm ** 4) + \
                                        np.abs(self.det ** 2)

        invariants_dict['sigma_minus'] = 0.5 * self.norm ** 2 - \
                                         np.sqrt(0.25 * self.norm ** 4) + \
                                         np.abs(self.det ** 2)

        return invariants_dict


# ==============================================================================
# errors
# ==============================================================================
class MT_Z_Error(Exception):
    pass


# ======================================================================
#                               TIPPER
# ======================================================================

class Tipper(object):
    """
    Tipper class --> generates a Tipper-object.

    Errors are given as standard deviations (sqrt(VAR))

    :param tipper_array: tipper array in the shape of [Tx, Ty]
                         *default* is None
    :type tipper_array: np.ndarray((nf, 1, 2), dtype='complex')


    :param tipper_err_array: array of estimated tipper errors
                             in the shape of [Tx, Ty].
                             Must be the same shape as tipper_array.
                             *default* is None
    :type tipper_err_array: np.ndarray((nf, 1, 2))


    :param freq: array of frequencies corresponding to the tipper elements.
                 Must be same length as tipper_array.
                 *default* is None
    :type freq: np.ndarray(nf)


    =============== ===========================================================
    Attributes      Description
    =============== ===========================================================
    freq            array of frequencies corresponding to elements of z
    rotation_angle  angle of which data is rotated by

    tipper          tipper array
    tipper_err       tipper error array
    =============== ===========================================================

    =============== ===========================================================
    Methods         Description
    =============== ===========================================================
    mag_direction   computes magnitude and direction of real and imaginary
                    induction arrows.
    amp_phase       computes amplitude and phase of Tx and Ty.
    rotate          rotates the data by the given angle
    =============== ===========================================================
    """

    def __init__(self, tipper_array=None, tipper_err_array=None,
                 freq=None):
        """
        initialize
        """
        self._logger = csamtpylog.get_csamtpy_logger(self.__class__.__name__)
        self._tipper = tipper_array
        self._tipper_err = tipper_err_array
        self._freq = freq

        self.rotation_angle = 0.
        if self.tipper is not None:
            self.rotation_angle = np.zeros((len(self.tipper)))

        self._amplitude = None
        self._amplitude_err = None
        self._phase = None
        self._phase_err = None

        self._mag_real = None
        self._mag_imag = None
        self._angle_real = None
        self._angle_imag = None
        self._mag_err = None
        self._angle_err = None

        if self._tipper is not None and self._freq is not None:
            self.compute_amp_phase()
            self.compute_mag_direction()

    # ==========================================================================
    # Define get/set and properties
    # ==========================================================================
    # ----freq----------------------------------------------------------
    @property
    def freq(self):
        return self._freq

    @freq.setter
    def freq(self, freq_arr):
        """
        Set the array of freq.

        :param freq_arr: array of frequnecies (Hz)
        :type freq_arr: np.ndarray(num_frequencies)
        """
        if freq_arr is not None:
            self._freq = np.array(freq_arr)

        if self._freq.size is not len(self.tipper):
            self._logger.info('length of freq list/array not correct' + \
                              ' (%ii instead of %ii)' % (self._freq.size, len(self.tipper)))
            return

        # for consistency recalculate amplitude and phase
        self.compute_amp_phase()

    # ---tipper--------------------------------------------------------------
    @property
    def tipper(self):
        return self._tipper

    @tipper.setter
    def tipper(self, tipper_array):
        """
        Set the attribute *tipper*

        :param tipper_array: tipper array in the shape of [Tx, Ty]
                             *default* is None
        :type tipper_array: np.ndarray((nf, 1, 2), dtype='complex')
        """

        # check to see if the new tipper array is the same shape as the old
        if self._tipper is not None and self._tipper.shape != tipper_array.shape:
            raise MT_Z_Error('Shape of new "tipper" array does not match old' + \
                             'new shape {0} != old shape {1}'.format(tipper_array.shape,
                                                                     self._tipper.shape) + \
                             '\n***Make new Tipper object***')
        if tipper_array is not None:
            if len(tipper_array.shape) == 3 and tipper_array.shape[1:3] == (1, 2):
                if tipper_array.dtype in ['complex', 'float', 'int']:
                    self._tipper = tipper_array

        # neeed to set the rotation angle such that it is an array
        if self.rotation_angle is float:
            self.rotation_angle = np.repeat(self.rotation_angle,
                                            len(self._tipper))

            # for consistency recalculate mag and angle
        self.compute_mag_direction()

        # for consistency recalculate amplitude and phase
        self.compute_amp_phase()

    # ----tipper error---------------------------------------------------------
    @property
    def tipper_err(self):
        return self._tipper_err

    @tipper_err.setter
    def tipper_err(self, tipper_err_array):
        """
        Set the attribute *tipper_err*.

        :param tipper_err_array: array of estimated tipper errors
                                 in the shape of [Tx, Ty].
                                 Must be the same shape as tipper_array.
                                 *default* is None
        :type tipper_err_array: np.ndarray((nf, 1, 2))
        """
        if self.tipper_err is not None and \
                (self._tipper_err.shape != tipper_err_array.shape):
            raise MT_Z_Error('Shape of new "tipper_err" array does not match old' + \
                             'new shape {0} != old shape {1}'.format(tipper_err_array.shape),
                             self._tipper_err.shape)

        # make sure the input array is of required shape
        if tipper_err_array is not None:
            if len(tipper_err_array.shape) == 3 and \
                            tipper_err_array.shape[1:3] == (1, 2):
                if tipper_err_array.dtype in ['float', 'int']:
                    self._tipper_err = tipper_err_array

                    assert self._tipper_err.shape == self._tipper.shape

        # for consistency recalculate mag and angle
        self.compute_mag_direction()

        # for consistency recalculate amplitude and phase
        self.compute_amp_phase()

    # ----amplitude and phase
    def compute_amp_phase(self):
        """
        Sets attributes:
    			* *amplitude*
    			* *phase*
    			* *amplitude_err*
    			* *phase_err*

        values for resistivity are in in Ohm m and phase in degrees.
        """

        if self.tipper is None:
            # logging.error( 'tipper array is None - cannot calculate rho/phi')
            # print 'tipper array is None - cannot calculate rho/phi'
            return None

        self._amplitude_err = None
        self._phase_err = None
        if self.tipper_err is not None:
            self._amplitude_err = np.zeros(self.tipper_err.shape)
            self._phase_err = np.zeros(self.tipper_err.shape)

        self._amplitude = np.abs(self.tipper)
        self._phase = np.rad2deg(np.angle(self.tipper))

        if self.tipper_err is not None:
            for idx_f in range(len(self.tipper)):
                for jj in range(2):
                    if self.tipper_err is not None:
                        if type(self.tipper) == np.ma.core.MaskedArray:
                            if self.tipper.mask[idx_f, 0, jj]:
                                continue
                        r_err, phi_err = MTcc.propagate_error_rect2polar(
                            np.real(self.tipper[idx_f, 0, jj]),
                            self.tipper_err[idx_f, 0, jj],
                            np.imag(self.tipper[idx_f, 0, jj]),
                            self.tipper_err[idx_f, 0, jj])

                        self.amplitude_err[idx_f, 0, jj] = r_err
                        self._phase_err[idx_f, 0, jj] = phi_err

    def set_amp_phase(self, r_array, phi_array):
        """
        Set values for amplitude(r) and argument (phi - in degrees).

        Updates the attributes:
    			* tipper
    			* tipper_err

        """

        if self.tipper is not None:

            tipper_new = copy.copy(self.tipper)

            if self.tipper.shape != r_array.shape:
                self._logger.error('Error - shape of "r" array does not match shape of ' + \
                                   'tipper array: %s ; %s' % (str(r_array.shape),
                                                              str(self.tipper.shape)))
                return

            if self.tipper.shape != phi_array.shape:
                self._logger.error('Error - shape of "phi" array does not match shape of ' + \
                                   'tipper array: %s ; %s' % (str(phi_array.shape),
                                                              str(self.tipper.shape)))
                return
        else:

            tipper_new = np.zeros(r_array.shape, 'complex')

            if r_array.shape != phi_array.shape:
                self._logger.error('Error - shape of "phi" array does not match shape ' + \
                                   'of "r" array: %s ; %s' % (str(phi_array.shape),
                                                              str(r_array.shape)))
                return

        # assert real array:
        if np.linalg.norm(np.imag(r_array)) != 0:
            self._logger.error('Error - array "r" is not real valued !')
            return
        if np.linalg.norm(np.imag(phi_array)) != 0:
            self._logger.error('Error - array "phi" is not real valued !')
            return

        for idx_f in range(len(r_array)):
            for jj in range(2):
                tipper_new[idx_f, 0, jj] = cmath.rect(r_array[idx_f, 0, jj],
                                                      math.radians(phi_array[idx_f, 0, jj]))

        self.tipper = tipper_new

        # for consistency recalculate amplitude and phase
        self.compute_amp_phase()
        self.compute_mag_direction()

    # ---------------------------------
    # properties
    @property
    def amplitude(self):
        return self._amplitude

    @property
    def phase(self):
        return self._phase

    @property
    def amplitude_err(self):
        return self._amplitude_err

    @property
    def phase_err(self):
        return self._phase_err

    # ----magnitude and direction----------------------------------------------
    def compute_mag_direction(self):
        """
        computes the magnitude and direction of the real and imaginary
        induction vectors.
        """

        if self.tipper is None:
            return None
        self._mag_real = np.sqrt(self.tipper[:, 0, 0].real ** 2 + \
                                 self.tipper[:, 0, 1].real ** 2)
        self._mag_imag = np.sqrt(self.tipper[:, 0, 0].imag ** 2 +
                                 self.tipper[:, 0, 1].imag ** 2)

        self._mag_err = None
        self._angle_err = None
        # get the angle, need to make both parts negative to get it into the
        # parkinson convention where the arrows point towards the conductor

        self._angle_real = np.rad2deg(np.arctan2(-self.tipper[:, 0, 1].real,
                                                 -self.tipper[:, 0, 0].real))

        self._angle_imag = np.rad2deg(np.arctan2(-self.tipper[:, 0, 1].imag,
                                                 -self.tipper[:, 0, 0].imag))

        ## estimate error: THIS MAYBE A HACK
        if self.tipper_err is not None:
            self._mag_err = np.sqrt(self.tipper_err[:, 0, 0] ** 2 + \
                                    self.tipper_err[:, 0, 1] ** 2)
            self._angle_err = np.rad2deg(np.arctan2(self.tipper_err[:, 0, 0],
                                                    self.tipper_err[:, 0, 1])) % 45

    def set_mag_direction(self, mag_real, ang_real, mag_imag, ang_imag):
        """
        computes the tipper from the magnitude and direction of the real
        and imaginary components.

        Updates tipper

        No error propagation yet
        """

        self.tipper[:, 0, 0].real = np.sqrt((mag_real ** 2 * np.arctan(ang_real) ** 2) / \
                                            (1 - np.arctan(ang_real) ** 2))

        self.tipper[:, 0, 1].real = np.sqrt(mag_real ** 2 / \
                                            (1 - np.arctan(ang_real) ** 2))

        self.tipper[:, 0, 0].imag = np.sqrt((mag_imag ** 2 * np.arctan(ang_imag) ** 2) / \
                                            (1 - np.arctan(ang_imag) ** 2))

        self.tipper[:, 0, 1].imag = np.sqrt(mag_imag ** 2 / \
                                            (1 - np.arctan(ang_imag) ** 2))
        # for consistency recalculate mag and angle
        self.compute_mag_direction()
        self.compute_amp_phase()

    @property
    def mag_real(self):
        return self._mag_real

    @property
    def mag_imag(self):
        return self._mag_imag

    @property
    def angle_real(self):
        return self._angle_real

    @property
    def angle_imag(self):
        return self._angle_imag

    @property
    def mag_err(self):
        return self._mag_err

    @property
    def angle_err(self):
        return self._angle_err

    # ----rotate---------------------------------------------------------------
    def rotate(self, alpha):
        """
        Rotate  Tipper array.

        Rotation angle must be given in degrees. All angles are referenced
        to geographic North=0, positive in clockwise direction.
        (Mathematically negative!)

        In non-rotated state, 'X' refs to North and 'Y' to East direction.

        Updates the attributes:
    			* *tipper*
    			* *tipper_err*
    			* *rotation_angle*

        """

        if self.tipper is None:
            self._logger.error('tipper array is "None" - I cannot rotate that')
            return

        # check for iterable list/set of angles - if so, it must have length 1
        # or same as len(tipper):
        if np.iterable(alpha) == 0:
            try:
                degreeangle = float(alpha % 360)
            except ValueError:
                self._logger.error('"Angle" must be a valid number (in degrees)')
                return

            # make an n long list of identical angles
            lo_angles = [degreeangle for ii in self.tipper]
        elif len(alpha) == 1:
            try:
                degreeangle = float(alpha % 360)
            except ValueError:
                self._logger.error('"Angle" must be a valid number (in degrees)')
                return
            # make an n long list of identical angles
            lo_angles = [degreeangle for ii in self.tipper]
        else:
            try:
                lo_angles = [float(ii % 360) for ii in alpha]
            except ValueError:
                self._logger.error('"Angles" must be valid numbers (in degrees)')
                return

        self.rotation_angle = np.array([(oldangle + lo_angles[ii]) % 360
                                        for ii, oldangle in enumerate(self.rotation_angle)])

        if len(lo_angles) != len(self.tipper):
            self._logger.error('Wrong number Number of "angles" - need %ii ' % (len(self.tipper)))
            self.rotation_angle = 0.
            return

        tipper_rot = copy.copy(self.tipper)
        tipper_err_rot = copy.copy(self.tipper_err)

        for idx_freq in range(len(tipper_rot)):
            angle = lo_angles[idx_freq]

            if self.tipper_err is not None:
                tipper_rot[idx_freq], tipper_err_rot[idx_freq] = \
                    MTcc.rotatevector_incl_errors(self.tipper[idx_freq, :, :],
                                                  angle,
                                                  self.tipper_err[idx_freq, :, :])
            else:
                tipper_rot[idx_freq], tipper_err_rot = \
                    MTcc.rotatevector_incl_errors(self.tipper[idx_freq, :, :],
                                                  angle)

        self.tipper = tipper_rot
        self.tipper_err = tipper_err_rot

        # for consistency recalculate mag and angle
        self.compute_mag_direction()

        # for consistency recalculate amplitude and phase
        self.compute_amp_phase()


# ------------------------
def correct4sensor_orientation(Z_prime, Bx=0, By=90, Ex=0, Ey=90,
                               Z_prime_error=None):
    """
    Correct a Z-array for wrong orientation of the sensors.

    Assume, E' is measured by sensors orientated with the angles
        E'x: a
        E'y: b

    Assume, B' is measured by sensors orientated with the angles
        B'x: c
        B'y: d

    With those data, one obtained the impedance tensor Z':
        E' = Z' * B'

    Now we define change-of-basis matrices T,U so that
        E = T * E'
        B = U * B'

    =>   T contains the expression of the E'-basis in terms of E
    (the standard basis)
    and  U contains the expression of the B'-basis in terms of B
    (the standard basis)
    The respective expressions for E'x-basis vector and E'y-basis
    vector are the columns of T.
    The respective expressions for B'x-basis vector and B'y-basis
    vector are the columns of U.

    We obtain the impedance tensor in default coordinates as:

    E' = Z' * B' => T^(-1) * E = Z' * U^(-1) * B
                 => E = T * Z' * U^(-1) * B
                 => Z = T * Z' * U^(-1)

    :param Z_prime: impedance tensor to be adjusted
    :dtype Z_prime: np.ndarray(num_freq, 2, 2, dtype='complex')


    :param Bx: orientation of Bx relative to geographic north (0)
				   *default* is 0
    :type Bx: float (angle in degrees)

    :param By:
    :type By: float (angle in degrees)
		         orientation of By relative to geographic north (0)
				 *default* is 90

    :param Ex: orientation of Ex relative to geographic north (0)
				   *default* is 0
    :type Ex: float (angle in degrees)

    :param Ey: orientation of Ey relative to geographic north (0)
				  *default* is 90
    :type Ey: float (angle in degrees)

    :param Z_prime_error: impedance tensor error (std)
            					 *default* is None
    :type Z_prime_error: np.ndarray(Z_prime.shape)

    :returns: adjusted impedance tensor
    :rtype: np.ndarray(Z_prime.shape, dtype='complex')

    :returns: impedance tensor standard deviation in
					default orientation
    :rtype: np.ndarray(Z_prime.shape, dtype='real')
    """
    try:
        if len(Z_prime.shape) != 2:
            raise
        if Z_prime.shape != (2, 2):
            raise

        if Z_prime.dtype not in ['complex', 'float', 'int']:
            raise

        Z_prime = np.matrix(Z_prime)

    except:
        raise MTex.MTpyError_inputarguments('ERROR - Z array not valid!' + \
                                            'Must be 2x2 complex array')

    if Z_prime_error is not None:
        try:
            if len(Z_prime_error.shape) != 2:
                raise
            if Z_prime_error.shape != (2, 2):
                raise

            if Z_prime_error.dtype not in ['float', 'int']:
                raise

        except:
            raise MTex.MTpyError_inputarguments('ERROR - Z-error array not' + \
                                                'valid! Must be 2x2 real array')

    T = np.matrix(np.zeros((2, 2)))
    U = np.matrix(np.zeros((2, 2)))

    dummy1 = cmath.rect(1, math.radians(Ex))

    T[0, 0] = np.real(dummy1)
    T[1, 0] = np.imag(dummy1)
    dummy2 = cmath.rect(1, math.radians(Ey))
    T[0, 1] = np.real(dummy2)
    T[1, 1] = np.imag(dummy2)

    dummy3 = cmath.rect(1, math.radians(Bx))
    U[0, 0] = np.real(dummy3)
    U[1, 0] = np.imag(dummy3)
    dummy4 = cmath.rect(1, math.radians(By))
    U[0, 1] = np.real(dummy4)
    U[1, 1] = np.imag(dummy4)

    try:
        z_arr = np.array(np.dot(T, np.dot(Z_prime, U.I)))
    except:
        raise MTex.MTpyError_inputarguments("ERROR - Given angles do not" + \
                                            "define basis for 2 dimensions - cannot convert Z'")

    z_err_arr = copy.copy(Z_prime_error)

    # TODO: calculate error propagation

    return z_arr, z_err_arr



# -*- coding: utf-8 -*-
"""
===============================================================================
    Copyright © 2021  Kouadio K.Laurent
    
    This file is part of pyCSAMT.
    
    pyCSAMT is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    pyCSAMT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with pyCSAMT.  If not, see <https://www.gnu.org/licenses/>.

===============================================================================  
.. _module-Strata::
    
    :synopsis:Metaclasses of geostrata . Takes conventional strata geocodes 
            look in _geocodes folder . will be deprecated soon
    
Created on Wed Nov 25 16:45:33 2020

@author: KouaoLaurent K. alias @Daniel03
"""

from pycsamt.ff.processing.callffunc import set_stratum_on_dict as setdico
try : 
    from pycsamt.utils._csamtpylog import csamtpylog
    _logger=csamtpylog.get_csamtpy_logger(__name__)
except :
    pass


traces_classes={}

class MetaStrata(type):
    "control the list of classes "
    def __init__(cls, name, bases, dict):
        type.__init__(cls, name, bases, dict)
        traces_classes[name]=cls
        
    def __repr__(self):
        return '<{0}.{1}: {2}, {3}>'.format(type(self).__name__, self.name,
                                            self.value, self.traces_classes)
        
class Strata(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstrata):
        dicstrata=setdico()[0]
        for keys , value in dicstrata.items():
            setattr(cls, keys, value)
    
class SedimentDeposits(metaclass=Strata) :
    """
    Parents of all metasediments classes
    """
    pass 

############### PROPRIETY META ################
    
class argillite(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys]

class AGLT(metaclass=argillite):
    pass

class alluvium(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class ALUV(metaclass=alluvium):
    pass

class amphibolite(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class AMP(metaclass=amphibolite):
    pass 

class anorthosite(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 

class ANS(metaclass=anorthosite):
    pass

class andesite(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys]
    
class ANT(metaclass=andesite):
    pass

class aplite(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class APL(metaclass=aplite):
    pass

class arkose(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class ARKS(metaclass=arkose):
    pass

class arenite(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
      
class ARNT(metaclass=arenite):
    pass

class altered_rock(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class ATRK(metaclass=altered_rock):
    pass

class augen_gneiss(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class AUGN (metaclass=augen_gneiss):
    pass

class banded_iron_formation(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 

class BIF(metaclass=banded_iron_formation):
    pass

class black_shale(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class BLSH(metaclass=black_shale):
    pass

class basalt(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class BLT(metaclass=basalt):
    pass

class breccia(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class BX(metaclass=breccia):
    pass
class calcarenite(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class CALR(metaclass=calcarenite):
    pass

class cavity(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class CAV(metaclass=cavity):
    pass

class carbonate_iron_formation(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class CBIF(metaclass=carbonate_iron_formation):
    pass

class carbonate_rock(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class CBRK(metaclass=carbonate_rock):
    pass

class carbonatite(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class CBT(metaclass=carbonatite):
    pass

class chalk(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class CHLK(metaclass=chalk):
    pass

class chert(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 

class CHRT(metaclass=chert):
    pass

class clast(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class CLAS(metaclass=clast):
    pass

class clast_supported_breccia(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class CLBX(metaclass=clast_supported_breccia):
    pass

class calcrete(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class CLCR(metaclass=calcrete):
    pass

class clay(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class CLY(metaclass=clay):
    pass

class conglomerate(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 

class CNGL(metaclass=conglomerate):
    pass

class coal(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
    
class COAL(metaclass=coal):
    pass

class colluvium(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
      
class COLV(metaclass=colluvium):
    pass

class calc_silicate_rock(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 
      
class CSRK(metaclass=calc_silicate_rock):
    pass

class carbonate_vein(metaclass=MetaStrata):
    def __new__(cls, name, bases, dicstratum):
        dicstratum=Strata.__dict__[cls.__name__]
        for keys, values  in dicstratum.items():
            if values =='':values=None
            setattr(cls,keys,values )
        return cls.__dict__[keys] 

class CVN(metaclass=carbonate_vein):
    pass



        

if __name__=='__main__':
    
    # print(amphibolite.name)
    # print(type(argillite))
    # print(Strata.amphibolite)
    print(amphibolite.color)
    # # # print(basalt.size)
    # print(dir(wood))
    # for keys, values in set_stratum_on_dict()[0].items():
    #     if keys =='argillite':
    #         print(values)



# -*- coding: utf-8 -*-
"""
Created on Sat Sep 26 20:30:41 2020

@author: KouaoLaurent alias  @Daniel03

Class : 
    **Agso ** . 
            Data of Geological Welllogs 

"""
import os , sys 
# import numpy as np
import warnings
from pycsamt.utils._csamtpylog import csamtpylog
import pandas as pd


_logger=csamtpylog.get_csamtpy_logger(__name__)
# _logger.basicConfig(filename='LogTest')

sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('.'))

class Agso (object):
    """
    Read Agso as pandas Series
    Geological conventional rocks and structurals handling. 
    
    staticsmethods :
    ----------------
        * _agso_ * : func 
            return a dataframe from AGS0 file 
        
        * _agso_on_dict_: func
            return a special dictionnay of element of AGSO file 
        
    :Example: :: 
        >>>  import os 
        >>>  path=os.path.join(os.environ ['pyCSAMT;],'pycsamt', 'geodrill',
                                           _geocodes'
        - Test with AGSO.csv file
        >>> geo=Agso._agso_on_dict_(set_agsoDataFrame=False, return_orientation="SERIES")  
        ... print(agf)
        >>> agf_struc=Agso._agso_(configuration_agso_filename=os.path.basename(os.path.join(path,"AGSO_STCODES.csv")))
                
       - Test test with "AGSO_STCODES
       >>>  geo=Agso._agso_on_dict_(set_agsoDataFrame=True, return_orientation="series", 
                                   agso_codefile=os.path.basename(os.path.join(path,"AGSO_STCODES.csv")))

       - extract element with their correspondance 
      >>>  dico_test={key:value for key , value in zip (geo["CODE"],geo['__DESCRIPTION'])}
                dico_lst=sorted(dico_test.items())
      ...  print(dico_lst)
        
    """
    @staticmethod
    def _agso_(configuration_agso_filename =None):
        """
        Returns
        -------
        df_ : object 
            Read pandas series of Agsofiles.

        """
        
        if configuration_agso_filename is None :

            # path =os.path.normpath(os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
            #                                     'geodrill','_geocodes','AGSO.csv'))
            path = os.path.join(os.path.abspath('.'), 'pycsamt',
                                        'geodrill', '_geocodes', 'AGSO.csv')
            # os.chdir(os.path.dirname(path))
            # fileagso =os.path.basename(path)
            # os.chdir(path)
            df_=pd.read_csv(path)
        elif configuration_agso_filename is not None :
            df_=pd.read_csv(configuration_agso_filename)
        
        return df_
    
    @staticmethod
    def _agso_on_dict_(**kwargs):#single_series_extraction=False,
        """

        Parameters
        ----------
        *set_agsoDataFrame* : Boolean  
                if False it will take the agso default filename 'AGSO.csv'

        * single_series_extraction* : Boolean  (deprecatedWarning)
                way to creat a dictionnary 
                Each column match its values  no including index 
                keys are columns if False it will return a dictionnary with index 
                
        * return_orientation * : str 
                specify the way to extract the dictionary 
                
        * agso_codefile : str , 
                agsoFilename input, it will call _agso_ staticmethod.
                and straighforwardly output the dictionnay . No need to call _agso_ methods
                default is AGSO.csv file.

        Returns
        -------
            dict, dictionary of agso elements
      
        """
        
        set_agsoDataFrame=kwargs.pop("set_agsoDataFrame", False)
        agso_codefile=kwargs.pop("agso_codefile", None)
        return_orientation=kwargs.pop("return_orientation", None)
        
        orients=["series", "index",
                                 "records", "split"]
        
        # files_contents=[ii for ii in os.listdir(os.path.getcwd) if os.path.isfile(ii)]
        dico={}
        
        if set_agsoDataFrame is False:
            agsoDataFrame=Agso._agso_()
        elif set_agsoDataFrame is True:
            if agso_codefile is not None :
                assert type(agso_codefile)==str , "Wrong Type ! Must be a string format"
                if not agso_codefile.endswith(".csv"):
                    _logger.error("Wrong Format of agsoFile ! Check your  _geocodes Folder !")
                    warnings.warn("Wrong agso Format ! must  choose "\
                                  "file amongs files on _geocode Folder!",category="error")#, \
                                  #format(",".join([ii for ii in os.listdir(os.getcwd()) if os.path.isfile(ii)])))
    
                agsoDataFrame=Agso._agso_(configuration_agso_filename=agso_codefile)
            elif agso_codefile is None :
                raise Exception('Must input at least the name of geocodefile ! '\
                                'or set argument "set_agsoDataFrame" to False'\
                                    ' to select the default agsoFile [AGSO.csv]. ')
                
        df=agsoDataFrame.copy()
        
        if return_orientation is None : # return the defaut orientation 
            dico=df.to_dict()
        else :
            try :
                if return_orientation.lower() in orients :
                    dico=df.to_dict(return_orientation.lower())
                elif return_orientation not in orients :
                    _logger.error("Must input the pandas series return_orientation."\
                                  " Some orientation are {0}".\
                                      format(','.join(["\{0}".format(jj) for jj in orients])))
                    print("Wrong orientation ! List of return_orientations are :[{0}]".\
                          format(','.join(["'{0}'".format(ii) for ii in orients])) )    
            except TypeError :
                pass
        ###################
        # this particular loop below are not necessary :
        # warnings.info(text=, )
        # if single_series_extraction is True :
        #     columnsNames=list(df.columns)
        #     for ii , val in enumerate(columnsNames):
        #         values=df[val]
        #         dico[val]=values
        # else :
            
        return dico
        
     
if __name__=="__main__":

    path=r'F:\OneDrive\Python\CodesExercices\ex_avgfiles\modules\_geocodes'
    os.chdir(path)
    
    #TEST with AGSO.csv file
    geo=Agso._agso_on_dict_(set_agsoDataFrame=False, return_orientation="SERies")
    
    # print(agf)
    # agf_struc=Agso._agso_(configuration_agso_filename=os.path.basename(os.path.join(path,"AGSO_STCODES.csv")))
    
    #test with "AGSO_STCODES
    # geo=Agso._agso_on_dict_(set_agsoDataFrame=True, return_orientation="series", 
    #                         agso_codefile=os.path.basename(os.path.join(path,"AGSO_STCODES.csv")))
    
    #EXTRACT ELEMENT WITH THEIR CORRESPONDANCE 
    dico_test={key:value for key , value in zip (geo["CODE"],geo['__DESCRIPTION'])}
    dico_lst=sorted(dico_test.items())
    # print(dico_lst)
    print(geo
          )
 
        
        
        
        
        
        
        
        
        


# -*- coding: utf-8 -*-
"""
.Console scripts. 
    $ python pycsamt write_occam2golden
    
    write model occam2d file to write YangBo Model data file
        called 'x,y,z' *.data file for post-processing.
        Script use output Occam 2D files such as Model, Iteration, Mesh and data file .
        
Created on Tue Apr  6 19:56:09 2021

@author: @Daniel03
"""
import os 
from pycsamt.modeling.occam2d import Iter2Dat as i2d
from pycsamt.gui.wrap_console_scripts import wrap_cscripts as wrs


default_kwargs = {'Path to occam2d files - str - ' : None , 
                  'Occam2d mesh file name - str - ' : None , 
                  'Occam2d model file - str - ': None , 
                  'Occam2d data file - str - ' : None, 
                  'Occam2d iteration file - str - ' : None, 
                  'Depth of investigation in meter** - float - ': 1000., 
                  'Scale  in [m|km] - str - ': 'm',
                  'Path to iter2dat files - str -':None,
                  'Output filename - str -': None , 
                  'Output negative depth - bool -' :True,
                  }

def main(): 
    """
    Can give input_kwargs to test consoles scripts.
    
    :Example:
        
        >>> input_kwargs ={'Path to occam2d files': r'C:/Users\Administrator\Desktop\ThesisImp\'\
        ...                       'occam2D\invers+files\inver_res\K1', 
        ...            'data':'OccamDataFile.dat',
        ...            'mesh' : 'Occam2DMesh',
        ...            'model' : 'Occam2DModel',
        ...            'iter':'ITER17.iter' }      
        >>> wrs_obj = wrs(default_kwargs =default_kwargs, input_kwargs =input_kwargs)
    
    """

    wrs_obj = wrs(default_kwargs =default_kwargs)
    
    for kwey, kwvalues in wrs_obj.sanitize_kwargs.items():
        if kwey.find('Path to occam2d') >=0 : path = kwvalues
        if kwey.find('mesh')>=0 : path_to_occam_mesh = kwvalues
        if kwey.find('data') >= 0 : path_to_occam_data = kwvalues 
        if kwey.find('iteration') >= 0 : path_to_occam_iter = kwvalues 
        if kwey.find('model') >=0: path_to_occam_model = kwvalues
        if kwey.find('investigation') >=0 :DOI = kwvalues
        if kwey.find('Scale') >= 0 : scale_output= kwvalues 
        if kwey.find('Output filename')>=0 : outputfilename = kwvalues

    occam_iter2dat_obj =i2d(mesh_fn=os.path.join(path, path_to_occam_mesh), 
                        iter_fn = os.path.join(path, path_to_occam_iter), 
                        model_fn =os.path.join(path, path_to_occam_model ), 
                        data_fn =os.path.join(path, path_to_occam_data))
    occam_iter2dat_obj.write_iter2dat_file(filename =outputfilename,
                                            scale=scale_output, 
                                            doi=DOI)
    
if __name__=='__main__':
    main()


# -*- coding: utf-8 -*-
"""
       .Specific exceptions for pyCSAMT.
    .Created on Wed Nov 18 17:17:13 2020

@author: @Daniel03

"""

#=================================================================
class pyCSAMTError_occam2d_plot(Exception): 
    pass 
class pyCSAMTError_occam2d(Exception): 
    pass 
class pyCSAMTError_occam2d_iter2dat(Exception): 
    pass

class pyCSAMTError_profile(Exception): 
    pass 

class pyCSAMTError_J(Exception):
    pass

class pyCSAMTError_site(Exception): 
    pass 

class MTpyError_Z(Exception):
    pass 

class MTpyError_inputarguments(Exception): 
    pass 

class pyCSAMTError_EDI(Exception):
    pass 

class pyCSAMTError_location(Exception): 
    pass 
class pyCSAMTError_well_building(Exception): 
    pass 

class pyCSAMTError_float(Exception):
    pass
class pyCSAMTError_frequency(Exception):
    pass

class pyCSAMTError_inputarguments(Exception):
    pass

class pyCSAMTError_Header(Exception):
    pass 

class pyCSAMTError_value(Exception):
    pass
class pyCSAMTError_azimuth(Exception): 
    pass 

class pyCSAMTError_ts_data(Exception):
    pass


class pyCSAMTError_config_file(Exception):
    pass


class pyCSAMTError_file_handling(Exception):
    pass


class pyCSAMTError_avg_file(Exception):
    pass

class pyCSAMTError_AvgData(Exception):
    pass


class pyCSAMTError_AVG(Exception):
    pass

class pyCSAMTError_plot_tip(Exception):
    pass 
class pyCSAMTError_plot(Exception): 
    pass 

class pyCSAMTError_Z(Exception):
    pass


class pyCSAMTError_Emag(Exception):
    pass


class pyCSAMTError_Hmag(Exception):
    pass

class pyCSAMTError_Ephz(Exception):
    pass

class pyCSAMTError_parameter_number(Exception):
    pass

class pyCSAMTError_processing(Exception):
    pass

class pyCSAMTError_module_import(Exception):
    pass

class pyCSAMTError_Phase(Exception):
    pass 

class pyCSAMTError_rho(Exception):
    pass

class pyCSAMTError_station(Exception):
    pass

class pyCSAMTError_structural(Exception): 
    pass

class pyCSAMTError_strata(Exception): 
    pass

class pyCSAMTError_plot_geoinputargument(Exception): 
    pass
class pyCSAMTError_geodrill_inputarguments(Exception): 
    pass

class pyCSAMTError_SQL(Exception): 
    pass
class pyCSAMTError_SQL_manager(Exception): 
    pass
class pyCSAMTError_SQL_interface(Exception): 
    pass
class pyCSAMTError_SQL_geoDataBase(Exception): 
    pass

class pyCSAMTError_SQL_update_geoinformation(Exception): 
    pass




